from __future__ import annotations

import attrs
import json
import logging
import pandas as pd
import pylsl as lsl
import typing as tp
from typing import ClassVar

from NaviNIBS.util.attrs import attrsAsDict
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util import exceptionToStr

from NaviNIBS.Navigator.Model.GenericCollection import GenericCollection, GenericCollectionDictItem

logger = logging.getLogger(__name__)


@attrs.define
class TriggerEvent:
    type: str
    time: pd.Timestamp = attrs.field(factory=pd.Timestamp.now)
    metadata: tp.Dict[str, tp.Any] = attrs.field(factory=dict)


@attrs.define
class TriggerSource(GenericCollectionDictItem[str]):
    type: ClassVar[str] = 'TriggerSource'
    _isEnabled: bool = True

    sigTriggered: Signal = attrs.field(init=False, factory=lambda: Signal((TriggerEvent,)))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @property
    def isEnabled(self):
        return self._isEnabled

    @isEnabled.setter
    def isEnabled(self, newIsEnabled: bool):
        if self._isEnabled == newIsEnabled:
            return
        self.sigItemAboutToChange.emit(self.key, ['isEnabled'])
        self._isEnabled = newIsEnabled
        self.sigItemChanged.emit(self.key, ['isEnabled'])

    def trigger(self, triggerEvt: TriggerEvent):
        if not self.isEnabled:
            logger.debug(f'Dropping event {triggerEvt} because trigger source is disabled')
            return
        if 'source' not in triggerEvt.metadata:
            triggerEvt.metadata['source'] = self.type
        logger.debug(f'Recorded trigger: {triggerEvt}')
        self.sigTriggered.emit(triggerEvt)

    def asDict(self) -> dict[str, tp.Any]:
        d = attrsAsDict(self)
        d['type'] = self.type
        return d

    @classmethod
    def fromDict(cls, d):
        type = d.pop('type')
        match(type):
            case LSLTriggerSource.type:
                return LSLTriggerSource(**d)
            case HotkeyTriggerSource.type:
                return HotkeyTriggerSource.fromDict(d)
            case _:
                raise NotImplementedError(f'Unexpected trigger source type: {type}')


@attrs.define
class LSLTriggerSource(TriggerSource):
    type: ClassVar[str] = 'LSLTriggerSource'
    _streamKey: tp.Optional[str] = None
    _fallbackTriggerSourceKey: str | None = None
    """
    If the primary LSL stream is not available, can fallback to another trigger source. The fallback source
    will be disabled when the primary is available, and enabled when the primary is not available.
    """
    _triggerEvents: tp.Optional[dict[str, tp.Optional[str]]] = None  # dict mapping of {eventValue: action} on which to trigger
    _triggerValueIsEpochID: bool = False
    _defaultAction: str = 'pulse'
    _minInterTriggerPeriod: float = 0.2  # ignore repeated triggers within this time

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @property
    def streamKey(self):
        return self._streamKey

    @streamKey.setter
    def streamKey(self, newKey: tp.Optional[str]):
        if self._streamKey == newKey:
            return
        self.sigItemAboutToChange.emit(self.key, ['streamKey'])
        self._streamKey = newKey
        self.sigItemChanged.emit(self.key, ['streamKey'])

    @property
    def fallbackTriggerSourceKey(self):
        return self._fallbackTriggerSourceKey

    @fallbackTriggerSourceKey.setter
    def fallbackTriggerSourceKey(self, newKey: tp.Optional[str]):
        if self._fallbackTriggerSourceKey == newKey:
            return
        self.sigItemAboutToChange.emit(self.key, ['fallbackTriggerSourceKey'])
        self._fallbackTriggerSourceKey = newKey
        self.sigItemChanged.emit(self.key, ['fallbackTriggerSourceKey'])

    @property
    def triggerEvents(self):
        return self._triggerEvents

    @triggerEvents.setter
    def triggerEvents(self, newEvents: tp.Optional[dict[str, tp.Optional[str]]]):
        if self._triggerEvents == newEvents:
            return
        self.sigItemAboutToChange.emit(self.key, ['triggerEvents'])
        self._triggerEvents = newEvents
        self.sigItemChanged.emit(self.key, ['triggerEvents'])

    @property
    def triggerValueIsEpochID(self):
        return self._triggerValueIsEpochID

    @triggerValueIsEpochID.setter
    def triggerValueIsEpochID(self, newValue: bool):
        if self._triggerValueIsEpochID == newValue:
            return
        self.sigItemAboutToChange.emit(self.key, ['triggerValueIsEpochID'])
        self._triggerValueIsEpochID = newValue
        self.sigItemChanged.emit(self.key, ['triggerValueIsEpochID'])

    @property
    def defaultAction(self):
        return self._defaultAction

    @property
    def minInterTriggerPeriod(self):
        return self._minInterTriggerPeriod

    def trigger(self, triggerEvt: TriggerEvent):
        if 'source' not in triggerEvt.metadata:
            triggerEvt.metadata['source'] = self.type + ' ' + self._streamKey
        super().trigger(triggerEvt=triggerEvt)


@attrs.define
class Hotkey(GenericCollectionDictItem[str]):
    _key: str
    """
    Literal key that will trigger hotkey (e.g. "." or "PgDn")
    """
    _action: str
    """
    Action to trigger when hotkey fires (e.g. "sample", "previous", or "next")
    """
    _keyboardDeviceID: tp.Optional[str] = None
    """
    Can specify a keyboard's device ID to only respond to keypresses from that device. Useful for situations
    like when using a presenter remote that generates common keypresses like '.' and 'esc', and you don't want
    triggers to be generated when pressing the same keys on a normal keyboard.

    If None, no device filter will be applied.
    """

    @property
    def action(self):
        return self._action

    @action.setter
    def action(self, newAction: str):
        if self._action == newAction:
            return
        self.sigItemAboutToChange.emit(self.key, ['action'])
        self._action = newAction
        self.sigItemChanged.emit(self.key, ['action'])

    @property
    def keyboardDeviceID(self):
        return self._keyboardDeviceID

    @keyboardDeviceID.setter
    def keyboardDeviceID(self, newID: tp.Optional[str]):
        if self._keyboardDeviceID == newID:
            return

        self.sigItemAboutToChange.emit(self.key, ['keyboardDeviceID'])
        self._keyboardDeviceID = newID
        self.sigItemChanged.emit(self.key, ['keyboardDeviceID'])


@attrs.define
class Hotkeys(GenericCollection[str, Hotkey]):
    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @classmethod
    def fromList(cls, itemList: list[dict[str, tp.Any]]) -> Hotkeys:
        items = {}
        for itemDict in itemList:
            items[itemDict['key']] = Hotkey.fromDict(itemDict)

        return cls(items=items)


@attrs.define
class HotkeyTriggerSource(TriggerSource):
    type: ClassVar[str] = 'HotkeyTriggerSource'

    _hotkeys: Hotkeys = attrs.field(factory=Hotkeys)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._hotkeys.sigItemsAboutToChange.connect(lambda *args: self.sigItemAboutToChange.emit(self.key, ['hotkeys']))
        self._hotkeys.sigItemsChanged.connect(lambda *args: self.sigItemChanged.emit(self.key, ['hotkeys']))

    @property
    def hotkeys(self):
        return self._hotkeys

    def asDict(self) -> dict[str, tp.Any]:
        d = attrsAsDict(self, exclude=['hotkeys'])
        d['type'] = self.type
        d['hotkeys'] = self.hotkeys.asList()
        return d

    @classmethod
    def fromDict(cls, d):
        if 'hotkeys' in d:
            d['hotkeys'] = Hotkeys.fromList(d['hotkeys'])
        return cls(**d)


@attrs.define
class TriggerReceiver:
    _key: str
    _minTimeBetweenEvents: tp.Optional[float] = None

    _lastTriggeredEvent: tp.Optional[TriggerEvent] = attrs.field(init=False, default=None)
    sigTriggered: Signal = attrs.field(init=False, factory=lambda: Signal((TriggerEvent,)))

    def __attrs_post_init__(self):
        pass

    @property
    def key(self):
        return self._key

    def trigger(self, event: TriggerEvent):
        if self._minTimeBetweenEvents is not None and self._lastTriggeredEvent is not None:
            timeBetweenEvents = (event.time - self._lastTriggeredEvent.time).total_seconds()
            if timeBetweenEvents < self._minTimeBetweenEvents:
                logger.debug(f'Event {event} occured too quickly after previous, ignoring.')
                return

        self._lastTriggeredEvent = event
        self.sigTriggered.emit(event)
        logger.debug(f'Triggered {event}')


@attrs.define
class TriggerRouter:
    """
    Handle routing of triggers to any of multiple trigger receivers, e.g. to handle cases where when the
    registration panel is in the foreground, triggers should record a new registration point, but if the
    navigation panel is in the foreground, triggers should record a new coil orientation sample.

    Note: this is implemented here for centralization, but is not persisted in session file(s)
    """
    _receivers: dict[str, TriggerReceiver] = attrs.field(init=False, factory=dict)
    _exclusiveTriggerStacks: dict[str, list[str]] = attrs.field(init=False, factory=dict)
    _nonexclusiveTriggerReceivers: dict[str, set[str]] = attrs.field(init=False, factory=dict)

    def __attrs_post_init__(self):
        pass

    def registerReceiver(self, receiver: TriggerReceiver):
        assert receiver.key not in self._receivers
        self._receivers[receiver.key] = receiver

    def subscribeToTrigger(self, receiver: TriggerReceiver, triggerKey: str, exclusive: bool):
        assert receiver.key in self._receivers

        logger.debug(f"TriggerReceiver '{receiver.key}' subscribing to '{triggerKey}' {'exclusively' if exclusive else 'non-exclusively'}")

        if exclusive:
            if triggerKey not in self._exclusiveTriggerStacks:
                self._exclusiveTriggerStacks[triggerKey] = list()
            self._exclusiveTriggerStacks[triggerKey].append(receiver.key)
        else:
            if triggerKey not in self._nonexclusiveTriggerReceivers:
                self._nonexclusiveTriggerReceivers[triggerKey] = set()
            self._nonexclusiveTriggerReceivers[triggerKey].add(receiver.key)

    def unsubscribeFromTrigger(self, receiver: TriggerReceiver, triggerKey: str, exclusive: bool):
        assert receiver.key in self._receivers
        if exclusive:
            if triggerKey in self._exclusiveTriggerStacks and receiver.key in self._exclusiveTriggerStacks[triggerKey]:
                self._exclusiveTriggerStacks[triggerKey].remove(receiver.key)
            else:
                pass  # already unsubscribed
        else:
            if triggerKey in self._nonexclusiveTriggerReceivers and receiver.key in self._nonexclusiveTriggerReceivers[triggerKey]:
                self._nonexclusiveTriggerReceivers[triggerKey].remove(receiver.key)
            else:
                pass  # already unsubscribed

    def _onSourceTriggered(self, triggerEvt: TriggerEvent):
        if triggerEvt.type in self._exclusiveTriggerStacks:
            if len(self._exclusiveTriggerStacks[triggerEvt.type]) > 0:
                receiverKey = self._exclusiveTriggerStacks[triggerEvt.type][-1]
                logger.debug(f'Sending exclusive trigger {triggerEvt.type} to {receiverKey}')
                try:
                    self._receivers[receiverKey].trigger(triggerEvt)
                except Exception as e:
                    logger.error(f'Problem while signaling exclusive trigger: \n{exceptionToStr(e)}')
        if triggerEvt.type in self._nonexclusiveTriggerReceivers:
            for receiverKey in self._nonexclusiveTriggerReceivers[triggerEvt.type]:
                self._receivers[receiverKey].trigger(triggerEvt)

    def connectToTriggerSource(self, source: TriggerSource):
        source.sigTriggered.connect(self._onSourceTriggered)

    def disconnectFromTriggerSource(self, source: TriggerSource):
        source.sigTriggered.disconnect(self._onSourceTriggered)


@attrs.define
class TriggerSources(GenericCollection[str, TriggerSource]):
    _triggerRouter: TriggerRouter = attrs.field(init=False, factory=TriggerRouter)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        for key, source in self.items():
            self._triggerRouter.connectToTriggerSource(source)

    @property
    def triggerRouter(self):
        return self._triggerRouter

    def setItem(self, item: TriggerSource):
        key = item.key
        if key in self._items:
            self._triggerRouter.disconnectFromTriggerSource(self[key])
        super().setItem(item=item)
        self._triggerRouter.connectToTriggerSource(item)

    @classmethod
    def fromList(cls, itemList: list[dict[str, tp.Any]]) -> TriggerSources:
        items = {}
        for itemDict in itemList:
            items[itemDict['key']] = TriggerSource.fromDict(itemDict)

        return cls(items=items)






import attrs
import logging
from math import ceil, log10, trunc, fmod
from qtpy import QtWidgets, QtCore, QtGui

from NaviNIBS.util.Signaler import Signal

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class _CustomQDial(QtWidgets.QDial):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def hasHeightForWidth(self) -> bool:
        return True

    def heightForWidth(self, width: int) -> int:
        return width


@attrs.define
class AngleDial:
    _value: float = 0.
    _resolution: float = 0.1
    _singleStep: float | None = None  # in degrees
    _pageStep: float | None = None  # in degrees
    _doInvert: bool = False
    """
    When doInvert is False, clockwise rotation from 0 increases the angle.
    """
    _offsetAngle: float = 0.
    """
    In degrees.
    When offset is zero, value=0 is pointing down.
    When offset is +90 (and doInvert is False), value=0 is pointing left.
    """
    _centerAngle: float = 180.
    """
    In degrees.
    When centerAngle is zero, values range from (-180, 180).
    When centerAngle is 180, values range from (0, 360).
    """

    _enabled: bool = True

    _movingValue: float | None = attrs.field(init=False, default=None)
    _dialStartedMove: bool = attrs.field(init=False, default=False)
    _qdial: _CustomQDial = attrs.field(init=False)
    _numericField: QtWidgets.QDoubleSpinBox = attrs.field(init=False)
    _container: QtWidgets.QWidget = attrs.field(init=False, factory=QtWidgets.QWidget)
    _maxDialHeight: int | None = 100

    _layout: QtWidgets.QBoxLayout | None = None

    sigValueMoved: Signal = attrs.field(init=False, factory=lambda: Signal((float,)))
    """
    Emitted frequentlny while dial is moved or while numeric field is edited.
    Connect to this if you want to respond immediately to any GUI changes.
    """
    sigValueChanged: Signal = attrs.field(init=False, factory=lambda: Signal((float,)))
    """
    Emitted less frequently while dial is moved, or after numeric field signals editingFinished.
    Connect to this if you want to only start responding after a change is finalized.
    """

    def __attrs_post_init__(self):
        if self._layout is None:
            self._layout = QtWidgets.QHBoxLayout()
            self._layout.setContentsMargins(0, 0, 0, 0)
        self._container.setLayout(self._layout)

        self._qdial = _CustomQDial()
        options = QtWidgets.QStyleOptionSlider()
        self._qdial.initStyleOption(options)

        self._qdial.setWrapping(True)
        self._qdial.setNotchesVisible(True)
        self._qdial.setRange(0, ceil(360/self._resolution) - 1)
        self._qdial.setTracking(False)
        self._qdial.valueChanged.connect(self._onDialValueChanged)
        self._qdial.sliderMoved.connect(self._onDialValueMoved)
        self._qdial.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        if self._maxDialHeight is not None:
            self._qdial.setMaximumHeight(self._maxDialHeight)
        self._layout.addWidget(self._qdial)

        self._numericField = QtWidgets.QDoubleSpinBox()
        # self._numericField.setKeyboardTracking(False)
        self._numericField.setRange(-180 + self._centerAngle, 180 + self._centerAngle - self._resolution)
        if self._singleStep is None:
            self._numericField.setSingleStep(self._resolution * 10.)
            self._qdial.setSingleStep(ceil(1/self._resolution))
        else:
            self._numericField.setSingleStep(self._singleStep)
            self._qdial.setSingleStep(ceil(self._singleStep / self._resolution))

        if self._pageStep is None:
            if self._singleStep is None:
                self._qdial.setPageStep(ceil(5/self._resolution))
            else:
                self._qdial.setPageStep(ceil(self._singleStep * 5. / self._resolution))
        else:
            self._qdial.setPageStep(self._pageStep / self._resolution)

        self._numericField.setDecimals(int(ceil(log10(1 / self._resolution))))
        self._numericField.setSuffix('°')
        self._numericField.valueChanged.connect(self._onNumericFieldValueChanged)
        self._numericField.editingFinished.connect(self._onNumericFieldEditingFinished)
        self._numericField.setWrapping(True)
        self._layout.addWidget(self._numericField)

        self._qdial.setValue(self._angleToQDialValue(self._value))
        self._numericField.setValue(self._value)

        self._layout.addStretch()

        if not self._enabled:
            self._enabled = True
            self.enabled = False

    @property
    def wdgt(self):
        return self._container

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val: float):
        if val == self._value:
            return
        logger.debug(f'Value changed from {self._value} to {val}')
        self._value = val
        self._qdial.setValue(self._angleToQDialValue(val))
        self._numericField.setValue(val)
        # TODO: maybe emit sigValueMoved too
        self.sigValueChanged.emit(val)

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, val: bool):
        if val == self._enabled:
            return
        self._enabled = val
        self._qdial.setEnabled(val)
        self._numericField.setEnabled(val)

    def _qdialValueToAngle(self, val: int) -> float:
        origVal = val
        val = val * self._resolution
        val -= self._offsetAngle
        if self._doInvert:
            val = 360 - val
        val = ((val + (self._centerAngle - 180)) % 360) + (self._centerAngle - 180)
        val = round(val / self._resolution) * self._resolution
        logger.debug(f'qdial value to angle: {origVal} -> {val}')
        return val

    def _angleToQDialValue(self, val: float | None) -> int | None:
        if val is None:
            return None
        origVal = val
        if self._doInvert:
            val = 360 - val
        val += self._offsetAngle
        val = fmod(val, 360)
        val /= self._resolution
        val = int(trunc(val))
        logger.debug(f'angle to qdial value: {origVal} -> {val}')
        return val

    def _onDialValueMoved(self, newVal: int):
        newVal = self._qdialValueToAngle(newVal)
        if newVal == self._movingValue:
            return
        if newVal == self._value and self._movingValue is None:
            return

        logger.debug(f'value moved: {newVal}')
        self._dialStartedMove = True
        self._movingValue = newVal
        self._numericField.setValue(newVal)
        self.sigValueMoved.emit(newVal)

    def _onNumericFieldValueChanged(self, newVal: float):
        if newVal == self._movingValue:
            return
        currentValueRounded = round(self._value, int(ceil(log10(1 / self._resolution))))
        if newVal == currentValueRounded and self._movingValue is None:
            return
        logger.debug(f'value moved: {newVal}')
        self._dialStartedMove = False
        self._movingValue = newVal
        self._qdial.setValue(self._angleToQDialValue(newVal))
        self.sigValueMoved.emit(newVal)

    def _onDialValueChanged(self, newDialVal: int):
        if not self._dialStartedMove and self._movingValue is not None and self._angleToQDialValue(self._movingValue) == newDialVal:
            return  # numeric field is being edited, don't finalize here
        self._movingValue = None
        if self._angleToQDialValue(self._value) == newDialVal:
            return

        newVal = self._qdialValueToAngle(newDialVal)

        logger.debug(f'value changed: {newVal} ({newDialVal})')
        self._value = newVal
        self._numericField.setValue(newVal)
        self.sigValueChanged.emit(newVal)

    def _onNumericFieldEditingFinished(self):
        newVal = self._numericField.value()
        self._movingValue = None
        if newVal == self._value:
            return
        logger.debug(f'value changed: {newVal}')
        self._value = newVal
        self._qdial.setValue(self._angleToQDialValue(newVal))
        self.sigValueChanged.emit(newVal)


if __name__ == '__main__':
    logger.setLevel(logging.DEBUG)
    app = QtWidgets.QApplication([])
    win = QtWidgets.QMainWindow()
    win.setCentralWidget(QtWidgets.QWidget())
    layout = QtWidgets.QFormLayout()
    win.centralWidget().setLayout(layout)
    dial = AngleDial(value=45)
    layout.addRow('Default', dial.wdgt)

    dial2 = AngleDial(offsetAngle=180, doInvert=True, value=45)
    layout.addRow('Offset 180°, inverted', dial2.wdgt)

    dial3 = AngleDial(offsetAngle=90, centerAngle=0, value=45)
    layout.addRow('Offset 90°, center 0°', dial3.wdgt)

    QtCore.QTimer.singleShot(1000, lambda: setattr(dial, 'value', -1e-14))

    win.show()
    app.exec_()



import attrs
from .attrs import attrsAsDict
from .numpy import array_equalish
import numpy as np
import pytest
import typing as tp


@attrs.define
class ExampleClass:
    aStr: str = 'test'
    anOptionalStr: tp.Optional[str] = None
    anOptionalArray: tp.Optional[np.ndarray] = attrs.field(factory=lambda: np.zeros((2, 2)))
    _aPrivateOptionalArray: tp.Optional[np.ndarray] = attrs.field(default=None)
    _aPrivateInt: int = 0


@pytest.fixture
def objA():
    return ExampleClass()


@pytest.fixture
def objB():
    return ExampleClass(aStr='b',
                        anOptionalStr='b',
                        anOptionalArray=np.zeros((5, 5)),
                        aPrivateOptionalArray=np.full((2,2), 3),
                        aPrivateInt=10)


def test_equality(objA, objB):
    assert objA == ExampleClass()

    assert objA != objB

    assert ExampleClass(anOptionalArray=np.zeros((3, 3))) !=\
           ExampleClass(anOptionalArray=np.full((3, 3), 1))


def test_asdict(objA, objB):

    with pytest.raises(ValueError):
        dictA1 = attrsAsDict(objA)

    dictA = attrsAsDict(objA, eqs=dict(anOptionalArray=array_equalish,
                                       aPrivateOptionalArray=array_equalish))

    assert len(dictA) == 0

    dictB = attrsAsDict(objB, eqs=dict(anOptionalArray=array_equalish,
                                       aPrivateOptionalArray=array_equalish))

    assert len(dictB) == 5











from __future__ import annotations

import typing as tp

import attrs
from qtpy import QtWidgets

from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.util.GUI.Dock import Dock


TS = tp.TypeVar('TS')   # TriggerSource class (e.g. LSLTriggerSource) referenced by a settings widget


@attrs.define(kw_only=True)
class TriggerSourceSettingsWidget(tp.Generic[TS]):
    _dockKey: str
    _title: str

    _dock: Dock = attrs.field(init=False)
    _wdgt: QtWidgets.QWidget = attrs.field(factory=QtWidgets.QWidget)
    _triggerSourceKey: str
    _session: Session = attrs.field(repr=False)

    def __attrs_post_init__(self):
        self._dock = Dock(
            name=self._dockKey + self._title,
            closable=False,
            title=self._title,
            affinities=[self._dockKey]
        )
        self._dock.addWidget(self._wdgt)

    @property
    def dock(self):
        return self._dock

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, newSession: Session):
        self._session = newSession
        self._onSessionSet()

    def _onSessionSet(self):
        pass

    @property
    def triggerSource(self) -> tp.Optional[TS]:
        if self.session is not None and self._triggerSourceKey in self.session.triggerSources:
            return self.session.triggerSources[self._triggerSourceKey]
        else:
            return None



import collections.abc
import typing as tp
import numpy as np
import pyvista as pv
from pyvista import _vtk
if pv.__version__ <= '0.39.1':
    from pyvista.utilities.helpers import vtk_id_list_to_array
else:
    from pyvista.core.utilities import vtk_id_list_to_array


def find_closest_point(dataset: pv.DataSet, point: tp.Iterable, n: int = 1) -> int:
    """
    Similar to pv.DataSet.find_closest_point, but monkey-patches a cache of the
    vtkPointLocator to speed up repeated calls. It turns out this construction
    is more expensive than the actual find.

    Note: the cache is not automatically reset by changes to the dataset, so may break
    (or give incorrect results) if dataset is changed.
    """

    if not isinstance(point, (np.ndarray, collections.abc.Sequence)) or len(point) != 3:
        raise TypeError("Given point must be a length three sequence.")
    if not isinstance(n, int):
        raise TypeError("`n` must be a positive integer.")
    if n < 1:
        raise ValueError("`n` must be a positive integer.")

    if hasattr(dataset, '_point_locator'):
        locator = dataset._point_locator
    else:
        locator = _vtk.vtkPointLocator()
        locator.SetDataSet(dataset)
        locator.BuildLocator()
        dataset._point_locator = locator

        # must also monkey-patch getstate to not break pickling
        dataset.__getstate__orig = dataset.__getstate__
        def __getstate__(self):
            state = self.__getstate__orig()
            for key in ('_point_locator', '__getstate__orig', '__getstate__'):
                del state[key]
            return state
        dataset.__getstate__ = __getstate__.__get__(dataset, pv.DataSet)


    if n > 1:
        id_list = _vtk.vtkIdList()
        locator.FindClosestNPoints(n, point, id_list)
        return vtk_id_list_to_array(id_list)
    return locator.FindClosestPoint(point)





import asyncio
import attrs
import json
import logging
import numpy as np
import pytransform3d.rotations as ptr
from skspatial.objects import Line, Plane, Vector
from qtpy import QtWidgets, QtGui, QtCore
import typing as tp
import qtawesome as qta

from NaviNIBS.Navigator.Model.Targets import Target
from NaviNIBS.Devices.ToolPositionsClient import ToolPositionsClient
from NaviNIBS.Navigator.GUI.CollectionModels.TargetsTableModel import TargetsTableModel, FullTargetsTableModel
from NaviNIBS.Navigator.Model.Session import Session, Tool
from NaviNIBS.Navigator.Model.Calculations import getClosestPointToPointOnMesh, calculateCoilToMRITransfFromTargetEntryAngle
from NaviNIBS.util import exceptionToStr
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.Transforms import applyTransform, invertTransform, composeTransform, concatenateTransforms, applyDirectionTransform, calculateRotationMatrixFromVectorToVector
from NaviNIBS.util.GUI.Icons import getIcon
from NaviNIBS.util.GUI.QDial import AngleDial
from NaviNIBS.util.GUI.QScrollContainer import QScrollContainer
from NaviNIBS.util.GUI.QMouseWheelAdjustmentGuard import preventAnnoyingScrollBehaviour

logger = logging.getLogger(__name__)


@attrs.define
class CoordinateWidget:

    _session: tp.Optional[Session] = attrs.field(default=None, repr=False)
    _target: Target | None = None
    _whichCoord: str = 'target'  # target or entry
    _whichOtherCoord: str = 'entry'  # entry or target
    _doAutoUpdateOtherCoord: bool = True
    """
    Whether to automatically update entry coordinate if we are changing target coordinate
    (e.g. to keep overall trajectory roughly perpendicular to scalp)
    """

    _onNewCoordRequested: tp.Callable[[], np.ndarray] | None = attrs.field(default=None)
    """
    Use this to specify a callback for user pressing button to set new coordinates (e.g. from a cursor in a view outside this widget).
    
    Callable should return ndarray vector of coordinates in world space.
    """
    # TODO: maybe make this async instead to allow for a request to prompt user for feedback, etc.
    _setCoordButtonLabel: str = 'Set coord'

    _wdgt: QtWidgets.QWidget = attrs.field(factory=QtWidgets.QWidget)
    _layout: QtWidgets.QFormLayout = attrs.field(init=False, factory=QtWidgets.QFormLayout)
    _coordInSysWdgts: dict[str, QtWidgets.QLabel] = attrs.field(init=False, factory=dict)
    _editCoordInSysWdgts: dict[str, QtWidgets.QToolButton] = attrs.field(init=False, factory=dict)
    _setCoordButton: QtWidgets.QPushButton | None = attrs.field(init=False, default=None)
    _autoUpdateOtherCoordCheckbox: QtWidgets.QCheckBox | None = attrs.field(init=False, default=None)

    def __attrs_post_init__(self):
        self._wdgt.setLayout(self._layout)

        if self._onNewCoordRequested is not None:
            self._setCoordButton = QtWidgets.QPushButton(self._setCoordButtonLabel)
            self._setCoordButton.clicked.connect(self._onSetCoordButtonClicked)
            self._layout.addRow(self._setCoordButton)
            self._autoUpdateOtherCoordCheckbox = QtWidgets.QCheckBox('')
            self._autoUpdateOtherCoordCheckbox.setChecked(self._doAutoUpdateOtherCoord)
            self._layout.addRow(f'Auto-update {self._whichOtherCoord} on set', self._autoUpdateOtherCoordCheckbox)

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, newSes: tp.Optional[Session]):
        self._session = newSes
        self._redraw()

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, newTarget: Target | None):
        if self._target is newTarget:
            return
        if self._target is not None:
            self._target.sigItemChanged.disconnect(self._onTargetItemChanged)

        self._target = newTarget

        if self._target is not None:
            self._target.sigItemChanged.connect(self._onTargetItemChanged)

        self._redraw()

    @property
    def coordAttrib(self):
        return self._whichCoord + 'Coord'

    @property
    def wdgt(self):
        return self._wdgt

    def _onTargetItemChanged(self, targetKey: str, attribsChanged: list[str] | None = None):
        if attribsChanged is None or self.coordAttrib in attribsChanged:
            self._redraw()

    def _onSetCoordButtonClicked(self, _: bool):
        assert self._onNewCoordRequested is not None
        assert self.target is not None
        newCoord = self._onNewCoordRequested()
        logger.info(f'Changing {self.target.key} {self.coordAttrib} to {newCoord}')
        setattr(self.target, self.coordAttrib, newCoord)
        if self._autoUpdateOtherCoordCheckbox is not None and self._autoUpdateOtherCoordCheckbox.isChecked():
            match self._whichOtherCoord:
                case 'entry':
                    self.target.autosetEntryCoord()
                case 'target':
                    raise NotImplementedError('Auto-updating target from entry not yet implemented')
                    # TODO: add support for this in Target class, similar to ``autosetEntryCoord``
                case _:
                    raise NotImplementedError

    def _onEditCoordClicked(self, coordKey: str):

        worldCoord = getattr(self._target, self.coordAttrib)

        logger.info(f'Editing {coordKey} {self.coordAttrib} coordinates for target {self._target.key}')

        if coordKey == 'World':
            coord = worldCoord
        else:
            coord = self._session.coordinateSystems[coordKey].transformFromWorldToThis(worldCoord)

        coordTxt = ', '.join([f'{c:.2f}' for c in coord])

        respTxt, ok = QtWidgets.QInputDialog.getText(
            self.wdgt,
            f'Edit coordinates',
            f'{coordKey} x, y, z',
            QtWidgets.QLineEdit.Normal,
            coordTxt)

        if not ok:
            # edit was cancelled
            return

        try:
            newCoord = np.fromstring(respTxt, sep=',')

            if len(newCoord) == 0:
                raise ValueError('No coordinates entered')

            elif len(newCoord) != 3:
                raise ValueError('Expected 3 coordinates')
        except Exception as e:
            logger.warning(f'Unable to parse response {respTxt}. Canceling edit.\n{exceptionToStr(e)}')
            return

        logger.info(f'User specified coordinate: {newCoord}')

        if coordKey == 'World':
            newCoord_world = newCoord
        else:
            newCoord_world = self._session.coordinateSystems[coordKey].transformFromThisToWorld(newCoord)

            logger.info(f'New coordinates in world space: {newCoord_world}')

        setattr(self._target, self.coordAttrib, newCoord_world)

        if self._autoUpdateOtherCoordCheckbox is not None and self._autoUpdateOtherCoordCheckbox.isChecked():
            match self._whichOtherCoord:
                case 'entry':
                    self.target.autosetEntryCoord()
                case 'target':
                    raise NotImplementedError('Auto-updating target from entry not yet implemented')
                    # TODO: add support for this in Target class, similar to ``autosetEntryCoord``
                case _:
                    raise NotImplementedError

    def _redraw(self):
        if self._session is None:
            for wdgtsDict in (self._coordInSysWdgts, self._editCoordInSysWdgts):
                for key, wdgt in wdgtsDict.items():
                    self._layout.removeWidget(wdgt)
                wdgtsDict.clear()
            if self._setCoordButton is not None:
                self._setCoordButton.setEnabled(False)
            return

        if self._target is None:
            for key, wdgt in self._coordInSysWdgts.items():
                wdgt.setText('')
            for key, wdgt in self._editCoordInSysWdgts.items():
                wdgt.setEnabled(False)
            if self._setCoordButton is not None:
                self._setCoordButton.setEnabled(False)
            return

        for key, wdgt in self._editCoordInSysWdgts.items():
            wdgt.setEnabled(True)

        if self._setCoordButton is not None:
            self._setCoordButton.setEnabled(True)

        assert 'World' not in self._session.coordinateSystems
        coordSysKeys = ['World'] + list(self._session.coordinateSystems.keys())
        for iKey, key in enumerate(coordSysKeys):
            if key not in self._coordInSysWdgts:
                wdgt = QtWidgets.QLabel()
                if key != 'World':
                    wdgt.setToolTip(self._session.coordinateSystems[key].description)
                self._coordInSysWdgts[key] = wdgt
                if True:
                    # add an edit button next to displayed coordinates
                    container = QtWidgets.QWidget()
                    self._layout.insertRow(iKey, key, container)
                    container.setLayout(QtWidgets.QHBoxLayout())
                    container.layout().setContentsMargins(0, 0, 0, 0)
                    container.layout().addWidget(wdgt)
                    container.layout().addStretch()

                    editWdgt = QtWidgets.QToolButton()
                    self._editCoordInSysWdgts[key] = editWdgt
                    editWdgt.setIcon(getIcon('mdi6.pencil'))
                    editWdgt.clicked.connect(lambda *args, coordKey=key: self._onEditCoordClicked(coordKey))
                    container.layout().addWidget(editWdgt)
                else:
                    # read-only, don't create an edit button
                    self._layout.insertRow(iKey, key, wdgt)

            worldCoord = getattr(self._target, self.coordAttrib)

            if key == 'World':
                coord = worldCoord
            else:
                coord = self._session.coordinateSystems[key].transformFromWorldToThis(worldCoord)

            # assume one decimal point is sufficient for any coordinate system
            # (works for integer or mm units, not m units...)
            if False:
                coordTxt = json.dumps(np.around(coord, decimals=1).tolist())
                # note: this sometimes has floating point precision errors such that many more decimal places are shown
            else:
                coordTxt = '[' + ', '.join([f'{c:.2f}' for c in coord]) + ']'

            self._coordInSysWdgts[key].setText(coordTxt)
            self._wdgt.updateGeometry()

        removeKeys = set(self._coordInSysWdgts.keys()) - set(coordSysKeys)
        for key in removeKeys:
            wdgt = self._coordInSysWdgts.pop(key)
            self._layout.removeWidget(wdgt)


@attrs.define(kw_only=True)
class EntryAnglesWidgets:
    _layout: QtWidgets.QFormLayout
    _session: tp.Optional[Session] = attrs.field(default=None, repr=False)
    _target: Target | None = None

    _pivotWdgt: QtWidgets.QComboBox = attrs.field(init=False)
    _angleRefWdgt: QtWidgets.QComboBox = attrs.field(init=False)
    _angleXWdgt: AngleDial = attrs.field(init=False)
    _angleYWdgt: AngleDial = attrs.field(init=False)

    def __attrs_post_init__(self):
        # TODO: maybe put all entry angle widgets into groupbox
        self._pivotWdgt = QtWidgets.QComboBox()
        self._pivotWdgt.addItems(['Coil', 'Entry', 'Target'])
        self._pivotWdgt.setCurrentIndex(2)
        self._pivotWdgt.currentIndexChanged.connect(self._onEntryAnglePivotChanged)
        self._layout.addRow('Entry angle pivot', self._pivotWdgt)

        self._angleRefWdgt = QtWidgets.QComboBox()
        self._angleRefWdgt.addItems(['Closest scalp to target'])
        self._angleRefWdgt.setCurrentIndex(0)
        self._angleRefWdgt.currentIndexChanged.connect(self._onEntryAngleRefChanged)
        self._layout.addRow('Entry angle relative to', self._angleRefWdgt)

        self._angleXWdgt = AngleDial(
            doInvert=False,
            offsetAngle=180,
            centerAngle=0,
        )
        self._angleXWdgt.sigValueChanged.connect(self._onEntryAngleXChangedFromGUI)
        self._layout.addRow('Entry angle X', self._angleXWdgt.wdgt)

        self._angleYWdgt = AngleDial(
            doInvert=False,
            offsetAngle=180,
            centerAngle=0
        )
        self._angleYWdgt.sigValueChanged.connect(self._onEntryAngleYChangedFromGUI)
        self._layout.addRow('Entry angle Y', self._angleYWdgt.wdgt)

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, target: Target | None):
        if self._target is target:
            return

        if self._target is not None:
            self._target.sigItemChanged.disconnect(self._onTargetItemChanged)

        self._target = target

        if self._target is not None:
            self._target.sigItemChanged.connect(self._onTargetItemChanged)

        self._refreshEntryAngleWidgets()

    @property
    def pivotWdgt(self):
        return self._pivotWdgt

    @property
    def angleRefWdgt(self):
        return self._angleRefWdgt

    @property
    def angleXWdgt(self):
        return self._angleXWdgt

    @property
    def angleYWdgt(self):
        return self._angleYWdgt

    def _onEntryAnglePivotChanged(self, index: int):
        self._refreshEntryAngleWidgets()

    def _onEntryAngleRefChanged(self, index: int):
        self._refreshEntryAngleWidgets()

    def _onEntryAngleXChangedFromGUI(self, newAngle: float):
        logger.info(f'Entry angle X changed to {newAngle} degrees')
        self._applyEntryAngleChangesToModel()

    def _onEntryAngleYChangedFromGUI(self, newAngle: float):
        logger.info(f'Entry angle Y changed to {newAngle} degrees')
        self._applyEntryAngleChangesToModel()

    def _getAngleRefVec(self) -> Vector | None:
        angleRef = self._angleRefWdgt.currentText()
        match angleRef:
            case 'Closest scalp to target':
                closestPt_skin = getClosestPointToPointOnMesh(
                    session=self._session,
                    whichMesh='skinSurf',
                    point_MRISpace=self._target.targetCoord)
                if closestPt_skin is None:
                    return None
                idealNormal = Vector(closestPt_skin - self._target.targetCoord)
                return idealNormal

            case _:
                raise NotImplementedError

    def _clearAndDisableEntryAngleWidgets(self):
        self._angleXWdgt.value = 0
        self._angleYWdgt.value = 0
        self._angleXWdgt.enabled = False
        self._angleYWdgt.enabled = False

    def _refreshEntryAngleWidgets(self):
        if self._target is None:
            self._clearAndDisableEntryAngleWidgets()
            return

        angleRefVec_MRI = self._getAngleRefVec()
        if angleRefVec_MRI is None:
            self._clearAndDisableEntryAngleWidgets()
            return
        logger.debug(f'angleRefVec_MRI: {angleRefVec_MRI}')

        coilToMRITransf = self._target.coilToMRITransf
        if coilToMRITransf is None:
            self._clearAndDisableEntryAngleWidgets()
            return

        angleRefVec_coil = applyDirectionTransform(invertTransform(coilToMRITransf), angleRefVec_MRI)
        logger.debug(f'angleRefVec_coil: {angleRefVec_coil}')
        coilDepthVec_coil = Vector([0, 0, 1])
        logger.debug(f'coilDepthVec_coil: {coilDepthVec_coil}')

        rot = calculateRotationMatrixFromVectorToVector(angleRefVec_coil, coilDepthVec_coil)  # TODO: check sign
        logger.debug(f'rot: {rot}')

        angleZ, angleX, angleY = ptr.extrinsic_euler_zxy_from_active_matrix(rot)
        logger.debug(f'Entry angle X: {np.rad2deg(angleX)}  Y: {np.rad2deg(angleY)}  Z: {np.rad2deg(angleZ)}')

        with self._angleXWdgt.sigValueChanged.disconnected(self._onEntryAngleXChangedFromGUI), \
             self._angleYWdgt.sigValueChanged.disconnected(self._onEntryAngleYChangedFromGUI):
            self._angleXWdgt.value = np.rad2deg(angleX)
            self._angleYWdgt.value = np.rad2deg(angleY)

    def _applyEntryAngleChangesToModel(self):

        if self._target is None:
            logger.warning('Can\'t apply entry angle changes: no target selected')
            return

        refVec = self._getAngleRefVec()
        if refVec is None:
            logger.warning("Can't apply entry angle changes: no reference vector")
            return

        refVec = refVec.unit()

        entryToTargetDist = np.linalg.norm(self._target.entryCoord - self._target.targetCoord)

        pivot = self._pivotWdgt.currentText()
        match pivot:
            case 'Coil':
                coilToPivotZOffset = 0
                pivotCoord_MRISpace = self._target.entryCoordPlusDepthOffset
            case 'Entry':
                coilToPivotZOffset = self._target.depthOffset
                pivotCoord_MRISpace = self._target.entryCoord
            case 'Target':
                coilToPivotZOffset = self._target.depthOffset + entryToTargetDist
                pivotCoord_MRISpace = self._target.targetCoord

            case _:
                raise NotImplementedError

        prepivotCoilCoord_MRISpace = pivotCoord_MRISpace + refVec * coilToPivotZOffset
        prepivotEntryCoord_MRISpace = prepivotCoilCoord_MRISpace - refVec * self._target.depthOffset
        prepivotTargetCoord_MRISpace = prepivotCoilCoord_MRISpace - refVec * (self._target.depthOffset + entryToTargetDist)

        prepivotTargetCoilToMRITransf = calculateCoilToMRITransfFromTargetEntryAngle(
            session=self._session,
            targetCoord=prepivotTargetCoord_MRISpace,
            entryCoord=prepivotEntryCoord_MRISpace,
            angle=self._target.angle,
            depthOffset=self._target.depthOffset,
            prevCoilToMRITransf=self._target.coilToMRITransf,
        )

        coilToPivotSpaceTransf = np.eye(4)
        coilToPivotSpaceTransf[2, 3] = coilToPivotZOffset

        angleX = np.deg2rad(self._angleXWdgt.value)
        angleY = np.deg2rad(self._angleYWdgt.value)

        pivotTransf = np.eye(4)
        pivotTransf[:3, :3] = ptr.active_matrix_from_extrinsic_euler_zxy(np.asarray([0, angleX, angleY]))  # TODO: double check signs

        pivotedCoilToMRITransf = concatenateTransforms([coilToPivotSpaceTransf, pivotTransf, invertTransform(coilToPivotSpaceTransf), prepivotTargetCoilToMRITransf])

        self._target.entryCoord = applyTransform(pivotedCoilToMRITransf, np.asarray([0, 0, -self._target.depthOffset]), doCheck=False)
        self._target.targetCoord = applyTransform(pivotedCoilToMRITransf, np.asarray([0, 0, -self._target.depthOffset - entryToTargetDist]), doCheck=False)

    def _onTargetItemChanged(self, targetKey: str, attribsChanged: list[str] | None = None):
        if attribsChanged is None or any(x in attribsChanged for x in
                                         ('targetCoord',
                                          'entryCoord',
                                          'depthOffset',
                                          'coilToMRITransf')):
            self._refreshEntryAngleWidgets()


@attrs.define(kw_only=True)
class EditTargetWidget:
    _session: Session = attrs.field(repr=False)

    _wdgt: QtWidgets.QWidget = attrs.field(factory=lambda: QtWidgets.QGroupBox('Edit target'))
    _scroll: QScrollContainer = attrs.field(init=False)

    _depthOffsetWdgt: QtWidgets.QDoubleSpinBox = attrs.field(init=False)

    _getNewTargetCoord: tp.Callable[[], tp.Tuple[float, float, float]] | None = attrs.field(default=None)
    """
    Use this to specify a callback for user pressing button to set new target coordinates (e.g. from a cursor in a view outside this widget).
    """
    _setTargetCoordButtonLabel: str = 'Set target coord'

    _getNewEntryCoord: tp.Callable[[], tp.Tuple[float, float, float]] | None = attrs.field(default=None)
    """
    Use this to specify a callback for user pressing button to set new entry  coordinates (e.g. from a cursor in a view outside this widget).
    """
    _setEntryCoordButtonLabel: str = 'Set entry coord'

    _targetComboBox: QtWidgets.QComboBox = attrs.field(init=False, factory=QtWidgets.QComboBox)
    _targetsModel: FullTargetsTableModel = attrs.field(init=False)

    _targetCoordWdgt: CoordinateWidget = attrs.field(init=False)
    _entryCoordWdgt: CoordinateWidget = attrs.field(init=False)

    _handleAngleWdgt: AngleDial = attrs.field(init=False)

    _entryAngleWdgts: EntryAnglesWidgets = attrs.field(init=False)

    _target: Target | None = attrs.field(init=False, default=None)

    _doTrackModelSelectedTarget: bool = True

    _disableWidgetsWhenNoTarget: list[QtWidgets.QWidget] = attrs.field(init=False)

    def __attrs_post_init__(self):
        outerLayout = QtWidgets.QVBoxLayout()
        self._wdgt.setLayout(outerLayout)
        outerLayout.setContentsMargins(0, 0, 0, 0)

        layout = QtWidgets.QFormLayout()
        self._scroll = QScrollContainer(innerContainerLayout=layout)
        self._scroll.scrollArea.setFrameShape(QtWidgets.QFrame.NoFrame)
        outerLayout.addWidget(self._scroll.scrollArea)

        self._scroll.scrollArea.setSizePolicy(QtWidgets.QSizePolicy.Minimum,
                                              QtWidgets.QSizePolicy.Preferred)

        self._targetsModel = FullTargetsTableModel(session=self._session)
        self._targetComboBox.setModel(self._targetsModel)
        self._targetComboBox.currentIndexChanged.connect(self._onTargetComboBoxCurrentIndexChanged)
        self._targetsModel.sigSelectionChanged.connect(self._onModelSelectionChanged)
        layout.addRow('Editing target:', self._targetComboBox)

        self._depthOffsetWdgt = QtWidgets.QDoubleSpinBox()
        preventAnnoyingScrollBehaviour(self._depthOffsetWdgt)
        # self._depthOffsetWdgt.setKeyboardTracking(False)
        self._depthOffsetWdgt.valueChanged.connect(self._onDepthOffsetChangedFromGUI)
        self._depthOffsetWdgt.setRange(-10, 1000)
        self._depthOffsetWdgt.setSingleStep(0.1)
        self._depthOffsetWdgt.setDecimals(1)
        self._depthOffsetWdgt.setSuffix(' mm')
        layout.addRow('Depth offset:', self._depthOffsetWdgt)

        self._entryCoordWdgt = CoordinateWidget(self._session,
                                                whichCoord='entry',
                                                whichOtherCoord='target',
                                                doAutoUpdateOtherCoord=False,
                                                wdgt=QtWidgets.QGroupBox('Entry coordinate'),
                                                onNewCoordRequested=self._getNewEntryCoord,
                                                setCoordButtonLabel=self._setEntryCoordButtonLabel)
        layout.addRow(self._entryCoordWdgt.wdgt)

        self._targetCoordWdgt = CoordinateWidget(self._session,
                                                 whichCoord='target',
                                                 whichOtherCoord='entry',
                                                 doAutoUpdateOtherCoord=True,
                                                 wdgt=QtWidgets.QGroupBox('Target coordinate'),
                                                 onNewCoordRequested=self._getNewTargetCoord,
                                                 setCoordButtonLabel=self._setTargetCoordButtonLabel)
        layout.addRow(self._targetCoordWdgt.wdgt)

        self._handleAngleWdgt = AngleDial(
            doInvert=True,
            offsetAngle=0,
            centerAngle=0
        )
        self._handleAngleWdgt.sigValueChanged.connect(self._onHandleAngleChangedFromGUI)
        layout.addRow('Handle angle', self._handleAngleWdgt.wdgt)

        self._entryAngleWdgts = EntryAnglesWidgets(
            layout=layout,
            session=self._session,
            target=self._target
        )
        # children widgets added to layout internally in constructor

        self._disableWidgetsWhenNoTarget = [
            self._depthOffsetWdgt,
            self._targetCoordWdgt.wdgt,
            self._entryCoordWdgt.wdgt,
            self._handleAngleWdgt.wdgt,
            self._entryAngleWdgts.pivotWdgt,
            self._entryAngleWdgts.angleRefWdgt,
            self._entryAngleWdgts.angleXWdgt.wdgt,
            self._entryAngleWdgts.angleYWdgt.wdgt
        ]

        self._targetComboBox.setCurrentIndex(-1)
        self._onTargetComboBoxCurrentIndexChanged(self._targetComboBox.currentIndex())  # enable/disable widgets

    @property
    def wdgt(self):
        return self._wdgt

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, target: Target | None):
        if self._target is target:
            return

        if self._target is not None:
            self._target.sigItemChanged.disconnect(self._onTargetItemChanged)

        self._target = target

        if self._target is not None:
            self._target.sigItemChanged.connect(self._onTargetItemChanged)

        # update children widgets

        for obj in (self._targetCoordWdgt, self._entryCoordWdgt, self._entryAngleWdgts):
            obj.target = self._target

        if target is None:
            self._targetComboBox.setCurrentIndex(-1)
            self._handleAngleWdgt.value = 0
            self._depthOffsetWdgt.setValue(0)
        else:
            self._targetComboBox.setCurrentIndex(self._targetsModel.getIndexFromCollectionItemKey(target.key))
            self._handleAngleWdgt.value = target.angle
            self._depthOffsetWdgt.setValue(target.depthOffset)

    def setEnabled(self, enabled: bool):
        self._wdgt.setEnabled(enabled)

    def _onModelSelectionChanged(self, changedKeys: list[str]):
        if not self._doTrackModelSelectedTarget:
            return

        # make sure current combobox item is in selection

        menuCurrentTargetIndex = self._targetComboBox.currentIndex()
        if menuCurrentTargetIndex != -1:
            try:
                menuCurrentTargetKey = self._targetsModel.getCollectionItemKeyFromIndex(menuCurrentTargetIndex)
            except IndexError:
                # model probably just changed to reduce number of items
                return  # TODO: determine better behavior to handle this instead of just ignoring

            if menuCurrentTargetKey not in changedKeys:
                return

            if self._targetsModel.getCollectionItemIsSelected(menuCurrentTargetKey):
                return

        firstSelectedTargetKey = None
        for target in self._session.targets.values():
            if target.isSelected:
                firstSelectedTargetKey = target.key
                break

        if firstSelectedTargetKey is None:
            self._targetComboBox.setCurrentIndex(-1)
        else:
            index = self._targetsModel.getIndexFromCollectionItemKey(firstSelectedTargetKey)
            if False:
                self._targetComboBox.setCurrentIndex(index)
            else:
                # run after short delay to prevent issue with nested model changes
                QtCore.QTimer.singleShot(0, lambda index=index: self._targetComboBox.setCurrentIndex(index))

    def _onTargetComboBoxCurrentIndexChanged(self, index: int):

        if index == -1:
            self.target = None
            for wdgt in self._disableWidgetsWhenNoTarget:
                wdgt.setEnabled(False)
            return

        else:
            for wdgt in self._disableWidgetsWhenNoTarget:
                wdgt.setEnabled(True)

        targetKey = self._targetsModel.getCollectionItemKeyFromIndex(index)
        target = self._session.targets[targetKey]

        self.target = target

        if not target.isSelected:
            if False:
                # newly selected target in combobox is not selected in model,
                # so assume user doesn't want edit widget to track model selection
                self._doTrackModelSelectedTarget = False
            else:
                pass  # temporarily disabled for now due to issue with over-aggressive disabling
                # of tracking model's selected target (perhaps due to signal emit order?)

    def _onTargetItemChanged(self, targetKey: str, attribsChanged: list[str] | None = None):
        if attribsChanged is None or 'angle' in attribsChanged:
            self._handleAngleWdgt.value = self.target.angle

        if attribsChanged is None or 'depthOffset' in attribsChanged:
            self._depthOffsetWdgt.setValue(self.target.depthOffset)

    def _onHandleAngleChangedFromGUI(self, newAngle: float):
        logger.info(f'Handle angle changed to {newAngle} degrees')
        if self.target is not None:
            if self.target.angle != newAngle:
                logger.debug(f'Changing handle angle for {self.target.key} from {self.target.angle} to {newAngle}')
                self.target.angle = newAngle  # TODO: maybe check if angle (accounting for rounding) actually changed

    def _onDepthOffsetChangedFromGUI(self, newDepth: float):
        logger.info(f'Depth offset changed to {newDepth} mm')
        if self.target is not None:
            if round(self.target.depthOffset, 1) != round(newDepth, 1):
                logger.debug(f'Changing depth offset for {self.target.key} from {self.target.depthOffset} to {newDepth}')
                self.target.depthOffset = newDepth




from __future__ import annotations

import asyncio

import attrs
import logging
import numpy as np
import pyvista as pv
import typing as tp
from typing import ClassVar

from . import PlotViewLayer
from NaviNIBS.util.pyvista import Actor, setActorUserTransform, concatenateLineSegments


logger = logging.getLogger(__name__)


Transform = np.ndarray


@attrs.define
class TargetingCrosshairsLayer(PlotViewLayer):
    _type: ClassVar[str]

    _targetOrCoil: str = 'target'

    _color: str = '#0000ff'
    _opacity: float = 0.5
    _radius: float = 15.
    _offsetRadius: float = 7.5
    _lineWidth: float = 4.

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._coordinator.sigCurrentTargetChanged.connect(lambda: self._queueRedraw(which='initCrosshair'))
        self._coordinator.sigCurrentCoilPositionChanged.connect(lambda: self._queueRedraw(which=['updatePositions', 'crosshairVisibility']))

    def _redraw(self, which: tp.Union[tp.Optional[str], tp.List[str, ...]] = None):
        super()._redraw(which=which)

        if not isinstance(which, str):
            # assume parent call above triggered appropriate redraws
            return

        if which == 'all':
            which = ['initCrosshair']
            self._redraw(which=which)
            return

        elif which == 'clearCrosshair':
            actorKey = self._getActorKey('crosshair')
            if actorKey in self._actors:
                actor = self._actors.pop(actorKey)
                with self._plotter.allowNonblockingCalls():
                    self._plotter.remove_actor(actor)
                    self._plotter.render()

        elif which == 'crosshairVisibility':
            actorKey = self._getActorKey('crosshair')
            if actorKey in self._actors:
                actor = self._actors[actorKey]

                doShow = self._canShow()

                if actor.GetVisibility() != doShow:
                    actor.SetVisibility(doShow)
                    self._plotter.render()

        elif which == 'initCrosshair':
            actorKey = self._getActorKey('crosshair')

            target = self._coordinator.currentTarget
            if target is None:
                if self._targetOrCoil == 'target':
                    # no active target, cannot plot
                    self._redraw(which='clearCrosshair')
                    return
                elif self._targetOrCoil == 'coil':
                    # use an estimated zOffset until we have a target
                    zOffset = -10.
                else:
                    raise NotImplementedError()
            else:
                # distance from bottom of coil to target (presumably in brain)
                zOffset = -1 * np.linalg.norm(target.entryCoordPlusDepthOffset - target.targetCoord)

            lines = self._getCrosshairLineSegments(radius=self._radius)

            offsetLines = self._getCrosshairLineSegments(radius=self._offsetRadius, zOffset=zOffset)

            depthLine = pv.lines_from_points(np.asarray([[0, 0, 0], [0, 0, zOffset]]))

            self._actors[actorKey] = self._plotter.addLineSegments(
                                                     concatenateLineSegments([lines, offsetLines, depthLine]),
                                                     name=actorKey,
                                                     color=self._color,
                                                     width=self._lineWidth,
                                                     opacity=self._opacity)

            self._redraw(which=['updatePositions', 'crosshairVisibility'])

        elif which == 'updatePositions':
            if not self._canShow():
                return

            actorKey = self._getActorKey('crosshair')
            if actorKey not in self._actors:
                self._redraw(which='initCrosshair')
                return

            actor = self._actors[actorKey]

            if self._plotInSpace != 'MRI':
                raise NotImplementedError()  # TODO: add necessary transforms for plotting in other spaces below

            with self._plotter.allowNonblockingCalls():
                if self._targetOrCoil == 'target':
                    currentTargetToMRITransform = self._coordinator.currentTarget.coilToMRITransf
                    setActorUserTransform(actor, currentTargetToMRITransform)

                elif self._targetOrCoil == 'coil':
                    currentCoilToMRITransform = self._coordinator.currentCoilToMRITransform
                    setActorUserTransform(actor, currentCoilToMRITransform)

                else:
                    raise NotImplementedError()

                self._plotter.render()

        else:
            raise NotImplementedError('Unexpected redraw which: {}'.format(which))

    def _canShow(self) -> bool:
        canShow = True
        hasTarget = self._coordinator.currentTarget
        hasCoil = self._coordinator.currentCoilToMRITransform is not None

        if self._plotInSpace != 'MRI':
            raise NotImplementedError()  # TODO: check that we have necessary info to convert to other coordinate spaces

        if self._targetOrCoil == 'target':
            if not hasTarget:
                canShow = False
        elif self._targetOrCoil == 'coil':
            if not hasCoil:
                canShow = False
        else:
            raise NotImplementedError()
        return canShow

    @classmethod
    def _getCircleLines(cls, radius: float, numPts: int = 300) -> pv.PolyData:
        points = np.zeros((numPts, 3))
        theta = np.linspace(0, 2 * np.pi, numPts)
        points[:, 0] = radius * np.cos(theta)
        points[:, 1] = radius * np.sin(theta)
        return pv.lines_from_points(points)

    @classmethod
    def _getCrosshairLineSegments(cls,
                                  radius: float,
                                  numPtsInCircle: int = 300,
                                  zOffset: float = 0.,
                                  ) -> pv.PolyData:
        circle = cls._getCircleLines(radius=radius, numPts=numPtsInCircle)
        relNotchLength = 0.2
        # TODO: check signs and directions
        topNotch = pv.lines_from_points(np.asarray([[0, radius, 0], [0, radius * (1 - relNotchLength), 0]]))
        if False:
            # show a longer line at bottom notch to indicate handle angle
            botNotch = pv.lines_from_points(
                np.asarray([[0, -radius * (1 + relNotchLength), 0], [0, -radius * (1 - relNotchLength), 0]]))
        else:
            # show a small inward-pointing triangle at bottom notch to indicate handle angle
            angle = 5  # in degrees
            x1 = np.sin(np.deg2rad(angle))*-radius
            y1 = np.cos(np.deg2rad(angle))*-radius
            botNotch = pv.lines_from_points(
                np.asarray([[x1, y1, 0], [0, -radius * (1 - relNotchLength), 0], [-x1, y1, 0]])
            )

        leftNotch = pv.lines_from_points(
            np.asarray([[-radius, 0, 0], [-radius * (1 - relNotchLength), 0, 0]]))
        rightNotch = pv.lines_from_points(np.asarray([[radius, 0, 0], [radius * (1 - relNotchLength), 0, 0]]))

        lines = concatenateLineSegments([circle, botNotch, topNotch, leftNotch, rightNotch])

        lines.points += np.asarray([[0, 0, zOffset]])

        return lines


@attrs.define
class TargetingTargetCrosshairsLayer(TargetingCrosshairsLayer):
    _type: ClassVar[str] = 'TargetingTargetCrosshairs'
    _targetOrCoil: str = 'target'

    def __attrs_post_init__(self):
        super().__attrs_post_init__()


@attrs.define
class TargetingCoilCrosshairsLayer(TargetingCrosshairsLayer):
    _type: ClassVar[str] = 'TargetingCoilCrosshairs'
    _targetOrCoil: str = 'coil'

    _color: str = '#00ff00'
    _lineWidth: float = 8.

    def __attrs_post_init__(self):
        super().__attrs_post_init__()



from __future__ import annotations

import asyncio
import attrs
import logging
from qtpy import QtWidgets, QtGui, QtCore
import typing as tp

from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.util.GUI.Dock import Dock, DockArea
from NaviNIBS.util.Signaler import Signal


logger = logging.getLogger(__name__)


@attrs.define()
class MainViewPanel:
    _key: str
    _session: tp.Optional[Session] = attrs.field(default=None, repr=False)

    _label: tp.Optional[str] = None
    _icon: tp.Optional[QtGui.QIcon] = attrs.field(default=None)
    _iconFn: tp.Optional[tp.Callable[..., QtGui.QIcon]] = attrs.field(default=None, repr=False)
    """
    To be set by subclass. 
    If iconFn is set, icon will be ignored and iconFn will be called to generate the icon.
    This allows regenerating the icon automatically, e.g. after color palette changes.
    """

    _wdgt: QtWidgets.QWidget = attrs.field(init=False, factory=QtWidgets.QWidget)
    _dockWdgt: Dock = attrs.field(init=False)

    _hasInitialized: bool = attrs.field(init=False, default=False)
    _isInitializing: bool = attrs.field(init=False, default=False)  # True while in the middle of finishing initializing

    _isShown: bool = attrs.field(init=False, default=False)
    sigPanelShown: Signal = attrs.field(init=False, factory=Signal)
    sigPanelHidden: Signal = attrs.field(init=False, factory=Signal)

    def __attrs_post_init__(self):
        logger.info(f'Initializing {self.__class__.__name__}')
        self._dockWdgt = Dock(name=self._key,
                              title=self.label,
                              affinities=['MainViewPanel'],
                              icon=self._icon,
                              iconFn=self._iconFn)
        self._dockWdgt.addWidget(self._wdgt)

        self._dockWdgt.sigHidden.connect(self.sigPanelHidden.emit)
        self._dockWdgt.sigShown.connect(self.sigPanelShown.emit)

        self.sigPanelShown.connect(self._onPanelShown)
        self.sigPanelHidden.connect(self._onPanelHidden)

        self._wdgt.setEnabled(False)
        self._wdgt.setVisible(False)

    @property
    def key(self):
        return self._key

    @property
    def label(self):
        return self._label if self._label is not None else self._key

    @property
    def wdgt(self):
        return self._wdgt

    @property
    def dockWdgt(self):
        return self._dockWdgt

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, newVal: tp.Optional[session]):
        if self._session is newVal:
            return
        self._session = newVal
        self._onSessionSet()

    @property
    def isVisible(self):
        return self._dockWdgt.isVisible()

    @property
    def hasInitialized(self):
        return self._hasInitialized

    @property
    def isInitializing(self):
        return self._isInitializing

    def _onSessionSet(self):
        logger.debug(f'{self.__class__.__name__} onSessionSet')
        pass  # to be implemented by subclass

    def _onPanelShown(self):
        logger.info(f'Panel {self.key} shown')
        self._isShown = True
        if not self._hasInitialized and self.canBeEnabled()[0]:
            self.finishInitialization()

    def _onPanelHidden(self):
        logger.info(f'Panel {self.key} hidden')
        self._isShown = False

    def canBeEnabled(self) -> tuple[bool, str | None]:
        """
        Should return (True, None) if panel can be enabled, and (False, '<Reason why cannot be enabled>') otherwise.
        The second output in the latter case should be a string explaining why the panel cannot be enabled,
        e.g. 'Registration must be completed before navigating'
        """
        return True, None  # can be implemented by subclass to indicate when panel is missing critical information and can't yet show anything useful

    def updateEnabled(self):
        if self.canBeEnabled()[0]:
            self.dockWdgt.setEnabled(True)
            if self.isVisible:
                if not self.hasInitialized and not self.isInitializing:
                    self.finishInitialization()
            else:
                self.wdgt.setEnabled(True)
            self.dockWdgt.label.setToolTip(self.label)
        else:
            self.wdgt.setEnabled(False)
            self.dockWdgt.setEnabled(False)
            self.dockWdgt.label.setToolTip(f'{self.label}\n({self.canBeEnabled()[1]})')

    def finishInitialization(self):
        if self._isInitializing:
            logger.warning(f"{self.key} is already finishing initializing, skipping.")
            return

        if self._hasInitialized:
            logger.warning(f"{self.key} finish initialization requested, but has already initialized. Not running again.")
            return

        canBeEnabled, reason = self.canBeEnabled()
        if not canBeEnabled:
            logger.warning(f"{self.key} finish initialization requested, but not yet ready to be enabled ({reason}). Skipping.")
            return

        self._isInitializing = True
        self._finishInitialization()
        self._isInitializing = False

        self._wdgt.setEnabled(True)
        self._wdgt.setVisible(True)
        self._hasInitialized = True

    def _finishInitialization(self):
        logger.debug(f'{self.__class__.__name__} _finishInitialization')
        pass  # can be implemented by subclass

    def _onPanelDeactivated(self):
        pass

    def close(self):
        self._dockWdgt.close()


import logging

import attrs
import typing as tp
import os
from qtpy import QtWidgets, QtGui, QtCore

from NaviNIBS.util.Signaler import Signal


logger = logging.getLogger(__name__)


@attrs.define(init=False, slots=False)
class QFileSelectWidget(QtWidgets.QWidget):

    _browseMode: str  # one of ('getSaveFileName', 'getOpenFilename', 'getExistingDirectory')

    _filepath: tp.Optional[str] = None
    _showRelativeTo: tp.Optional[str] = None
    _showRelativePrefix: tp.Optional[str] = None  # if showing relative to path, can show this prefix in displayed QLineEdit to make origin of rel path clear, e.g. '[NaviNIBS]'
    _extFilters: tp.Optional[str] = None
    _browseCaption: tp.Optional[str] = None

    _textWidget: QtWidgets.QLineEdit = attrs.field(init=False)
    _browseBtn: QtWidgets.QPushButton = attrs.field(init=False)

    sigFilepathChanged: Signal = attrs.field(init=False, factory=lambda: Signal((str,)))

    def __init__(self, *args, parent: tp.Optional[QtWidgets.QWidget] = None, **kwargs):
        super().__init__(parent=parent)
        self.__attrs_init__(*args, **kwargs)

    def __attrs_post_init__(self):
        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self._textWidget = QtWidgets.QLineEdit()
        self._textWidget.setReadOnly(True)
        layout.addWidget(self._textWidget)

        self._browseBtn = QtWidgets.QPushButton('Browse')
        self._browseBtn.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self._browseBtn.clicked.connect(lambda _: self.browse())
        layout.addWidget(self._browseBtn)

        self.sigFilepathChanged.connect(lambda _: self._updateFilepathDisplay())

        self._updateFilepathDisplay()

    def browse(self):
        if self._browseCaption is None:
            browseCaption = ''  # TODO: do mode-dependent auto caption (e.g. 'Select save path')
        else:
            browseCaption = self._browseCaption

        if self._extFilters is None:
            extFilters = ''  # TODO: check if necessary
        else:
            extFilters = self._extFilters

        prevFilepath = self._filepath
        if prevFilepath is None and self._showRelativeTo is not None:
            prevFilepath = self._showRelativeTo

        logger.info('Showing browse dialog')
        if self._browseMode == 'getSaveFileName':
            newFilepath, _ = QtWidgets.QFileDialog.getSaveFileName(self,
                                                                   browseCaption,
                                                                   prevFilepath,
                                                                   extFilters)
        elif self._browseMode == 'getOpenFilename':
            newFilepath, _ = QtWidgets.QFileDialog.getOpenFileName(self,
                                                                   browseCaption,
                                                                   prevFilepath,
                                                                   extFilters)
        elif self._browseMode == 'getExistingDirectory':
            assert self._extFilters is None
            newFilepath, _ = QtWidgets.QFileDialog.getExistingDirectory(self,
                                                                   browseCaption,
                                                                   prevFilepath)
        else:
            raise NotImplementedError('Unexpected browseMode: {}'.format(self._browseMode))

        if len(newFilepath) == 0:
            logger.info('Browse cancelled')
            return

        newFilepath = os.path.normpath(newFilepath)

        logger.info('Browsed to filepath: {}'.format(newFilepath))
        self.filepath = newFilepath

    @property
    def filepath(self):
        if self._filepath is None:
            return None
        else:
            # convert to always use non-windows file separators for display
            return os.path.normpath(self._filepath)

    @filepath.setter
    def filepath(self, newFilepath: str | None):
        if self._filepath is None and newFilepath is None:
            return
        if self._filepath is not None and newFilepath is not None and \
                os.path.normpath(newFilepath) == os.path.normpath(self._filepath):
            return
        if newFilepath is not None:
            newFilepath = os.path.normpath(newFilepath)
        logger.info('Filepath changed from {} to {}'.format(self._filepath, newFilepath))
        self._filepath = newFilepath
        self.sigFilepathChanged.emit(newFilepath)

    @property
    def showRelativeTo(self):
        if self._showRelativeTo is None:
            return None
        else:
            # convert to always use non-windows file separators for display
            return self._showRelativeTo.replace('\\', '/')

    @showRelativeTo.setter
    def showRelativeTo(self, newPath: tp.Optional[str]):
        if newPath is not None:
            newPath = os.path.normpath(newPath)
        if self._showRelativeTo == newPath:
            return
        self._showRelativeTo = newPath
        self._updateFilepathDisplay()

    @property
    def showRelativePrefix(self):
        return self._showRelativePrefix

    @showRelativePrefix.setter
    def showRelativePrefix(self, newPrefix: tp.Optional[str]):
        if self._showRelativePrefix == newPrefix:
            return
        self._showRelativePrefix = newPrefix
        self._updateFilepathDisplay()

    def _updateFilepathDisplay(self):
        if self._filepath is None:
            displayPath = ''
        else:
            if self._showRelativeTo is not None:
                displayPath = os.path.relpath(self.filepath, self.showRelativeTo)  # TODO: confirm working as intended
                if self._showRelativePrefix is not None:
                    filesep = '\\' if '\\' in displayPath else '/'
                    displayPath = self._showRelativePrefix + filesep + displayPath
            else:
                displayPath = self._filepath
        self._textWidget.setText(displayPath)
        # TODO: scroll cursor to end of text widget (to show filename, effectively hide nested parent folders first)


from __future__ import annotations

import asyncio

import attrs
from datetime import datetime
import logging
import numpy as np
import os
import pathlib
import pyvista as pv
import qtawesome as qta
from qtpy import QtWidgets, QtGui, QtCore
import shutil
import typing as tp

from . import MainViewPanel
from NaviNIBS.Devices.ToolPositionsClient import ToolPositionsClient
from NaviNIBS.Navigator.GUI.Widgets.CollectionTableWidget import DigitizedLocationsTableWidget
from NaviNIBS.Navigator.GUI.Widgets.TrackingStatusWidget import TrackingStatusWidget
from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.Navigator.Model.DigitizedLocations import DigitizedLocation
from NaviNIBS.Navigator.Model.Tools import CoilTool, CalibrationPlate
from NaviNIBS.util import makeStrUnique
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.GUI.Icons import getIcon
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.pyvista import Actor, setActorUserTransform
from NaviNIBS.util.pyvista import DefaultBackgroundPlotter, RemotePlotterProxy
from NaviNIBS.util.Transforms import applyTransform, invertTransform, transformToString, stringToTransform, estimateAligningTransform, concatenateTransforms


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@attrs.define
class DigitizeLocationsPanel(MainViewPanel):
    _key: str = 'Digitize'
    _icon: QtGui.QIcon = attrs.field(init=False, factory=lambda: getIcon('mdi6.dots-hexagon'))
    _surfKey: str = 'skinSurf'

    _trackingStatusWdgt: TrackingStatusWidget = attrs.field(init=False)

    _sampleLocationBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _clearSampleLocationBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _deleteLocationBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _newLocationFromPointerBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _plotter: DefaultBackgroundPlotter = attrs.field(init=False)
    _actors: tp.Dict[str, tp.Optional[Actor]] = attrs.field(init=False, factory=dict)
    _positionsClient: tp.Optional[ToolPositionsClient] = attrs.field(init=False, default=None)
    _tblWdgt: DigitizedLocationsTableWidget = attrs.field(init=False)

    finishedAsyncInit: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def canBeEnabled(self) -> tuple[bool, str | None]:
        if self.session is None:
            return False, 'No session set'
        if not self.session.MRI.isSet:
            return False, 'No MRI set'
        if not self.session.headModel.isSet:
            return False, 'No head model set'
        if self.session.tools.subjectTracker is None:
            return False, 'No active subject tracker configured'
        if self.session.tools.pointer is None:
            return False, 'No active pointer tool configured'
        if not self.session.subjectRegistration.isRegistered:
            return False, 'Subject not registered'
        return True, None

    def _finishInitialization(self):
        super()._finishInitialization()

        self._positionsClient = ToolPositionsClient()
        self._positionsClient.sigLatestPositionsChanged.connect(lambda: self._redraw(which=['pointerPosition', 'sampleBtn']))

        self._wdgt.setLayout(QtWidgets.QHBoxLayout())

        sidebar = QtWidgets.QWidget()
        sidebar.setLayout(QtWidgets.QVBoxLayout())
        sidebar.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Expanding)
        self._wdgt.layout().addWidget(sidebar)

        self._trackingStatusWdgt = TrackingStatusWidget(session=self.session,
                                                        hideToolTypes=[CoilTool, CalibrationPlate])
        sidebar.layout().addWidget(self._trackingStatusWdgt.wdgt)

        samplesBox = QtWidgets.QGroupBox('Digitized locations')
        samplesBox.setLayout(QtWidgets.QVBoxLayout())
        sidebar.layout().addWidget(samplesBox)

        btnContainer = QtWidgets.QWidget()
        btnContainer.setLayout(QtWidgets.QGridLayout())
        samplesBox.layout().addWidget(btnContainer)

        btn = QtWidgets.QPushButton('Import montage')
        btn.clicked.connect(self._onImportMontageBtnClicked)
        btnContainer.layout().addWidget(btn, 0, 0)

        btn = QtWidgets.QPushButton('Sample location')
        btn.clicked.connect(self._onSampleLocationBtnClicked)
        btn.setEnabled(False)
        self._sampleLocationBtn = btn
        btnContainer.layout().addWidget(btn, 1, 0)

        btn = QtWidgets.QPushButton('Clear sampled location')
        btn.clicked.connect(self._onClearSampledLocationBtnClicked)
        btn.setEnabled(False)
        btnContainer.layout().addWidget(btn, 2, 0)
        # TODO: change this to "clear sampled locations" when multiple selected
        self._clearSampleLocationBtn = btn

        btn = QtWidgets.QPushButton('Delete row')
        btn.clicked.connect(self._onDeleteLocationBtnClicked)
        btn.setEnabled(False)
        btnContainer.layout().addWidget(btn, 3, 0)
        # TODO: change this to "clear sampled locations" when multiple selected
        self._deleteLocationBtn = btn

        self._tblWdgt = DigitizedLocationsTableWidget()
        self._tblWdgt.sigSelectionChanged.connect(self._onSelectedLocationsChanged)
        samplesBox.layout().addWidget(self._tblWdgt.wdgt)

        sidebar.layout().addStretch()

        self._plotter = DefaultBackgroundPlotter()
        self._wdgt.layout().addWidget(self._plotter)

        asyncio.create_task(asyncTryAndLogExceptionOnError(self._finishInitialization_async))

        if self.session is not None:
            self._onPanelInitializedAndSessionSet()

    async def _finishInitialization_async(self):
        if isinstance(self._plotter, RemotePlotterProxy):
            await self._plotter.isReadyEvent.wait()

        self._plotter.enable_depth_peeling(4)

        self._redraw(which='all')

        self.finishedAsyncInit.set()

    def _onSessionSet(self):
        super()._onSessionSet()
        # TODO: connect relevant session changed signals to _redraw calls

        if self._hasInitialized:
            self._onPanelInitializedAndSessionSet()

    def _onPanelInitializedAndSessionSet(self):
        self.session.headModel.sigDataChanged.connect(lambda which: self._redraw(which='initSurf'))
        self.session.digitizedLocations.sigItemsChanged.connect(self._onLocationsChanged)
        self.session.subjectRegistration.sigTrackerToMRITransfChanged.connect(lambda: self._redraw(which=[
            'initSampledLocations', 'initSubjectTracker', 'initPointer', 'pointerPosition']))

        self._trackingStatusWdgt.session = self.session

        self._tblWdgt.session = self.session
        self._tblWdgt.currentRow = 0

        self._redraw(which='all')

    def _onLocationsChanged(self, dlKeys: list[str], attribs: tp.Optional[list[str]] = None):
        if attribs is None or any(attrib in attribs for attrib in ('sampledCoord', 'color')):
            self._redraw(which='initSampledLocations')
        # TODO: add support for redrawing plannedCoords as needed

    def _onSelectedLocationsChanged(self, selectedIndices: list[int]):
        numSelLocs = len(self._tblWdgt.selectedCollectionItemKeys)
        if numSelLocs > 0:
            if numSelLocs == 1:
                self._deleteLocationBtn.setText('Delete row')
            else:
                self._deleteLocationBtn.setText('Delete rows')
            self._deleteLocationBtn.setEnabled(True)
        else:
            self._deleteLocationBtn.setText('Delete row')
            self._deleteLocationBtn.setEnabled(False)



    def _onSelectedLocationsChanged(self, selKeys: list[str]):
        if len(selKeys) == 0:
            self._clearSampleLocationBtn.setEnabled(False)
            self._clearSampleLocationBtn.setText('Clear sampled location')
        else:
            numSelLocationsWithSamples = sum(self.session.digitizedLocations[key].sampledCoord is not None for key in selKeys if key is not None)

            if numSelLocationsWithSamples == 0:
                self._clearSampleLocationBtn.setEnabled(False)
                self._clearSampleLocationBtn.setText('Clear sampled location')
                self._deleteLocationBtn.setEnabled(False)
                self._deleteLocationBtn.setText('Delete row')
            elif numSelLocationsWithSamples == 1:
                self._clearSampleLocationBtn.setEnabled(True)
                self._clearSampleLocationBtn.setText('Clear sampled location')
                self._deleteLocationBtn.setEnabled(True)
                self._deleteLocationBtn.setText('Delete row')
            else:
                self._clearSampleLocationBtn.setEnabled(True)
                self._clearSampleLocationBtn.setText('Clear sampled locations')
                self._deleteLocationBtn.setEnabled(True)
                self._deleteLocationBtn.setText('Delete rows')

            self._redraw(which='initSampledLocations')  # TODO: redraw just previously and currently selected points instead of all

    def _currentTblLocKey(self) -> tp.Optional[str]:
        return self._tblWdgt.currentCollectionItemKey

    def _getPointerCoordInMRISpace(self) -> tp.Optional[np.ndarray]:
        pointerToCameraTransf = self._positionsClient.getLatestTransf(self.session.tools.pointer.trackerKey, None)
        subjectTrackerToCameraTransf = self._positionsClient.getLatestTransf(self.session.tools.subjectTracker.trackerKey, None)
        subjectTrackerToMRITransf = self.session.subjectRegistration.trackerToMRITransf

        if pointerToCameraTransf is None or subjectTrackerToCameraTransf is None or subjectTrackerToMRITransf is None:
            logger.warning('Tried to sample, but do not have valid positions. Returning.')
            return None

        pointerCoord_MRISpace = applyTransform([self.session.tools.pointer.toolToTrackerTransf,
                                                       pointerToCameraTransf,
                                                       invertTransform(subjectTrackerToCameraTransf),
                                                       subjectTrackerToMRITransf],
                                               np.zeros((3,)),
                                               doCheck=False)

        return pointerCoord_MRISpace

    def _onImportMontageBtnClicked(self, checked: bool):
        filepath, _ = QtWidgets.QFileDialog.getOpenFileName(self._wdgt,
                                                               'Select montage file to import',
                                                               os.path.dirname(self.session.filepath),
                                                               'xml (*.xml)')

        if len(filepath) == 0:
            logger.info('Import cancelled')
            return

        self.session.digitizedLocations.loadFromXML(filepath)

    def _onSampleLocationBtnClicked(self):

        pointerCoord_MRISpace = self._getPointerCoordInMRISpace()
        if pointerCoord_MRISpace is None:
            return

        logger.info(f'Digitized location in MRI space: {pointerCoord_MRISpace}')

        locKey = self._currentTblLocKey()
        if locKey is None:
            # placeholder new row selected
            locKey = makeStrUnique(baseStr=f'Loc_{len(self.session.digitizedLocations) + 1}',
                                   existingStrs=self.session.digitizedLocations.keys(),
                                   delimiter='_')
            loc = DigitizedLocation(key=locKey,
                                    sampledCoord=pointerCoord_MRISpace)
            self.session.digitizedLocations.addItem(loc)
        else:
            self.session.digitizedLocations[locKey].sampledCoord = pointerCoord_MRISpace

        if True:
            currentRow = self._tblWdgt.currentRow
            if currentRow == self._tblWdgt.rowCount - 1:
                # already at end of table
                # TODO: auto-advance to prompt about aligning
                pass
            else:
                # advance to next fiducial in table
                self._tblWdgt.currentRow += 1

    def _onClearSampledLocationBtnClicked(self, checked: bool):
        for key in self._tblWdgt.selectedCollectionItemKeys:
            if key is not None:
                self.session.digitizedLocations[key].sampledCoord = None

    def _onDeleteLocationBtnClicked(self, checked: bool):
        selKeys = [key for key in self._tblWdgt.selectedCollectionItemKeys if key is not None]
        if len(selKeys) > 0:
            self.session.digitizedLocations.deleteItems(selKeys)

    def _redraw(self, which: tp.Union[str, tp.List[str,...]]):

        if isinstance(self._plotter, RemotePlotterProxy) and not self._plotter.isReadyEvent.is_set():
            # plotter not yet ready
            return

        logger.debug('redraw {}'.format(which))

        if not self.isVisible:
                return

        if isinstance(which, list):
            for subWhich in which:
                self._redraw(which=subWhich)
            return

        if which == 'all':
            which = ['initSurf', 'initSubjectTracker', 'initPointer', 'initSampledLocations',
                     'sampleBtn']
            self._redraw(which=which)
            return

        elif which == 'initSurf':
            actorKey = 'surf'
            self._actors[actorKey] = self._plotter.add_mesh(mesh=getattr(self.session.headModel, self._surfKey),
                                                            color='#d9a5b2',
                                                            opacity=0.8,  # TODO: make GUI-configurable
                                                            name=actorKey)

        elif which == 'initSubjectTracker':
            subjectTracker = self.session.tools.subjectTracker
            doShowSubjectTracker = self.session.subjectRegistration.trackerToMRITransf is not None \
                                and subjectTracker is not None \
                                and subjectTracker.trackerSurf is not None
            actorKey = 'subjectTracker'
            if not doShowSubjectTracker:
                if actorKey in self._actors:
                    self._plotter.remove_actor(self._actors[actorKey])
                    self._actors.pop(actorKey)

            else:
                self._actors[actorKey] = self._plotter.addMesh(mesh=self.session.tools.subjectTracker.trackerSurf,
                                                               color=self.session.tools.subjectTracker.trackerColor,
                                                               defaultMeshColor='#aaaaaa',
                                                               opacity=0.6,
                                                               name=actorKey)
                self._redraw(which='subjectTrackerPosition')

        elif which in ('sampleBtn',):
            pointer = self.session.tools.pointer
            subjectTracker = self.session.tools.subjectTracker

            allowSampling = False
            if pointer is not None and subjectTracker is not None:
                allowSampling = not any(self._positionsClient.getLatestTransf(key, None) is None for key in (pointer.trackerKey, subjectTracker.trackerKey))

            if self._sampleLocationBtn.isEnabled() != allowSampling:
                self._sampleLocationBtn.setEnabled(allowSampling)

        elif which in ('initPointer', 'pointerPosition'):
            pointer = self.session.tools.pointer
            subjectTracker = self.session.tools.subjectTracker

            doShowPointer = self.session.subjectRegistration.trackerToMRITransf is not None \
                            and pointer is not None \
                            and (pointer.trackerSurf is not None or pointer.toolSurf is not None) \
                            and subjectTracker is not None

            if not doShowPointer:
                for toolOrTracker in ('tool', 'tracker'):
                    actorKey = 'pointer' + '_' + toolOrTracker
                    if actorKey in self._actors:
                        self._plotter.remove_actor(self._actors[actorKey])
                        self._actors.pop(actorKey)
                return

            if which == 'initPointer':
                for toolOrTracker in ('tool', 'tracker'):
                    actorKey = 'pointer' + '_' + toolOrTracker
                    tool = self.session.tools.pointer
                    actorSurf = getattr(tool, toolOrTracker + 'Surf')
                    meshColor = getattr(tool, toolOrTracker + 'Color')
                    if actorSurf is None:
                        if actorKey in self._actors:
                            self._plotter.remove_actor(self._actors[actorKey])
                            self._actors.pop(actorKey)
                        continue
                    self._actors[actorKey] = self._plotter.addMesh(mesh=actorSurf,
                                                                   color=meshColor,
                                                                   defaultMeshColor='#999999',
                                                                   opacity=0.6,
                                                                   name=actorKey)
                self._redraw(which='pointerPosition')

            elif which == 'pointerPosition':
                for toolOrTracker in ('tool', 'tracker'):
                    actorKey = 'pointer' + '_' + toolOrTracker

                    if actorKey not in self._actors:
                        # assume this was because we don't have enough info to show
                        continue

                    pointerToCameraTransf = self._positionsClient.getLatestTransf(pointer.trackerKey, None)
                    subjectTrackerToCameraTransf = self._positionsClient.getLatestTransf(subjectTracker.trackerKey, None)

                    if pointerToCameraTransf is None or subjectTrackerToCameraTransf is None:
                        # don't have valid info for determining pointer position relative to head tracker
                        if self._actors[actorKey].GetVisibility():
                            self._actors[actorKey].VisibilityOff()
                        continue

                    if not self._actors[actorKey].GetVisibility():
                        self._actors[actorKey].VisibilityOn()

                    if toolOrTracker == 'tool':
                        pointerStlToSubjectTrackerTransf = concatenateTransforms([
                            pointer.toolStlToToolTransf,
                            pointer.toolToTrackerTransf,
                            pointerToCameraTransf,
                            invertTransform(subjectTrackerToCameraTransf)
                        ])
                    elif toolOrTracker == 'tracker':
                        pointerStlToSubjectTrackerTransf = concatenateTransforms([
                            pointer.trackerStlToTrackerTransf,
                            pointerToCameraTransf,
                            invertTransform(subjectTrackerToCameraTransf)
                        ])
                    else:
                        raise NotImplementedError()

                    setActorUserTransform(
                        self._actors[actorKey],
                        concatenateTransforms([
                            pointerStlToSubjectTrackerTransf,
                            self.session.subjectRegistration.trackerToMRITransf
                        ])
                    )
                    self._plotter.render()

            else:
                raise NotImplementedError()

        elif which == 'subjectTrackerPosition':
            actorKey = 'subjectTracker'
            if actorKey not in self._actors:
                # subject tracker hasn't been initialized, maybe due to missing information
                return

            with self._plotter.allowNonblockingCalls():
                setActorUserTransform(
                    self._actors[actorKey],
                    self.session.subjectRegistration.trackerToMRITransf @ self.session.tools.subjectTracker.trackerStlToTrackerTransf
                )

        elif which == 'initSampledLocations':

            actorKeys = ('sampledLocations_unselected', 'sampledLocations_selected')

            sampledLocKeys = [key for key, loc in self.session.digitizedLocations.items() if loc.sampledCoord is not None]
            doShowSampledLocs = len(sampledLocKeys) > 0

            if not doShowSampledLocs:
                # no sampled locations (yet)
                with self._plotter.allowNonblockingCalls():
                    for actorKey in actorKeys:
                        if actorKey in self._actors:
                            self._plotter.remove_actor(actorKey)
                            self._actors.pop(actorKey)
                return

            labels = []
            locsToPlot = {key: loc for key, loc in self.session.digitizedLocations.items()
                          if loc.sampledCoord is not None}
            coords = np.full((len(locsToPlot), 3), np.nan)

            selectedKeys = self._tblWdgt.selectedCollectionItemKeys

            for iLoc, (label, loc) in enumerate(locsToPlot.items()):
                labels.append(label)
                coords[iLoc, :] = loc.sampledCoord

            locColor = list(locsToPlot.values())[0].color
            # TODO: add support for per-electrode colors instead of using same for all
            # (will probably require multiple actors, at least one per unique color)

            unselectedLabels = [label for label in labels if label not in selectedKeys]
            unselectedCoords = coords[np.isin(labels, unselectedLabels), :]

            if len(unselectedCoords) > 0:
                self._actors[actorKeys[0]] = self._plotter.add_points(
                    name=actorKeys[0],
                    points=unselectedCoords,
                    color=locColor,
                    point_size=25,
                    render_points_as_spheres=True,
                    reset_camera=False,
                    render=True
                )
            else:
                actorKey = actorKeys[0]
                with self._plotter.allowNonblockingCalls():
                    if actorKey in self._actors:
                        self._plotter.remove_actor(actorKey)
                        self._actors.pop(actorKey)

            selectedLocColor = '#0096FF'
            selectedLabels = [label for label in labels if label in selectedKeys]
            selectedCoords = coords[np.isin(labels, selectedKeys), :]

            if len(selectedCoords) > 0:
                self._actors[actorKeys[1]] = self._plotter.add_point_labels(
                    name=actorKeys[1],
                    points=selectedCoords,
                    labels=selectedLabels,
                    point_color=selectedLocColor,
                    text_color=selectedLocColor,
                    point_size=25,
                    font_size=15,
                    always_visible=True,
                    shape=None,
                    render_points_as_spheres=True,
                    reset_camera=False,
                    render=True
                )
            else:
                actorKey = actorKeys[1]
                with self._plotter.allowNonblockingCalls():
                    if actorKey in self._actors:
                        self._plotter.remove_actor(actorKey)
                        self._actors.pop(actorKey)

        else:
            raise NotImplementedError('Unexpected redraw key: {}'.format(which))

        with self._plotter.allowNonblockingCalls():
            self._plotter.render()




from __future__ import annotations

import asyncio

import attrs
from datetime import datetime
import logging
import os
import pathlib
import pyqtgraph.dockarea.Container as pgdc
import qtawesome as qta
from qtpy import QtWidgets, QtGui, QtCore
import shutil
import typing as tp
from typing import TYPE_CHECKING

from NaviNIBS import __version__
from NaviNIBS.Navigator.GUI.ViewPanels.MainViewPanelWithDockWidgets import MainViewPanelWithDockWidgets
from NaviNIBS.Navigator.GUI.EditWindows.ImportSessionWindow import ImportSessionWindow
from NaviNIBS.util import exceptionToStr
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.GUI.Dock import Dock, DockArea, TContainer
from NaviNIBS.util.GUI.ErrorDialog import raiseErrorDialog
from NaviNIBS.util.GUI.Icons import getIcon
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.Navigator.Model.Addons import Addon, installPath as addonBaseInstallPath

if TYPE_CHECKING:
    from NaviNIBS.Navigator.GUI.NavigatorGUI import NavigatorGUI


logger = logging.getLogger(__name__)


@attrs.define(kw_only=True)
class ManageSessionPanel(MainViewPanelWithDockWidgets):
    _navigatorGUI: NavigatorGUI

    _icon: QtGui.QIcon = attrs.field(init=False, factory=lambda: getIcon('mdi6.form-select'))

    _autosavePeriod: float = 60  # in sec

    _inProgressBaseDir: tp.Optional[str] = None
    _rootDockArea: DockArea | None = attrs.field(init=False, default=None) # used for tabSaveBtn
    _newSessionBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _tabSaveBtn: QtWidgets.QPushButton | None = attrs.field(init=False, default=None)
    _importBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _saveBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _saveBtnShowsDirty: bool | None = attrs.field(init=False, default=None)
    _saveShortcut: QtWidgets.QShortcut = attrs.field(init=False)
    _saveToFileBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _saveToDirBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _closeBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _addAddonBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _fileDW: Dock = attrs.field(init=False)
    _fileContainer: QtWidgets.QWidget = attrs.field(init=False)
    _infoDW: Dock = attrs.field(init=False)
    _settingsContainer: QtWidgets.QWidget = attrs.field(init=False)
    _infoContainer: QtWidgets.QWidget = attrs.field(init=False)
    _appearanceContainer: QtWidgets.QWidget = attrs.field(init=False)
    _infoWdgts: tp.Dict[str, QtWidgets.QLineEdit] = attrs.field(init=False, factory=dict)
    _themeDropdown: QtWidgets.QComboBox = attrs.field(init=False)
    _fontSizeField: QtWidgets.QSpinBox = attrs.field(init=False)
    _importSessionWindow: ImportSessionWindow | None = attrs.field(init=False, default=None)

    _autosaveTask: asyncio.Task = attrs.field(init=False)

    sigAboutToFinishLoadingSession: Signal = attrs.field(init=False, factory=lambda: Signal((Session,)))
    sigLoadedSession: Signal = attrs.field(init=False, factory=lambda: Signal((Session,)))
    sigClosedSession: Signal = attrs.field(init=False, factory=lambda: Signal((Session,)))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        title = 'File'
        dock = Dock(
            name=self._key + title,
            closable=False,
            title=title,
            affinities=[self._key])
        self._fileDW = dock
        container = QtWidgets.QWidget()
        self._fileContainer = container
        container.setLayout(QtWidgets.QVBoxLayout())
        container.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.MinimumExpanding)
        dock.addWidget(container)
        dock.setStretch(1, 10)
        self._wdgt.addDock(dock, position='left')

        btn = QtWidgets.QPushButton(icon=getIcon('mdi6.file-plus'), text='New session')
        self._newSessionBtn = btn
        btn.clicked.connect(lambda checked: self._createNewSession())
        container.layout().addWidget(btn)

        btn = QtWidgets.QPushButton(icon=getIcon('mdi6.folder-open'), text='Load session')
        btn.clicked.connect(lambda checked: self.loadSession())
        container.layout().addWidget(btn)

        btn = QtWidgets.QPushButton(icon=getIcon('mdi6.folder-plus-outline'), text='Copy from session...')
        btn.clicked.connect(lambda checked: self.importSession())
        container.layout().addWidget(btn)
        self._importBtn = btn

        btn = QtWidgets.QPushButton(icon=getIcon('mdi6.file-restore'), text='Recover in-progress session')
        btn.clicked.connect(lambda checked: self._recoverSession())
        container.layout().addWidget(btn)

        btn = QtWidgets.QPushButton(icon=getIcon('mdi6.clipboard-file'), text='Clone session')
        btn.clicked.connect(lambda checked: self._cloneSession())
        container.layout().addWidget(btn)

        container.layout().addSpacing(10)

        btn = QtWidgets.QPushButton(icon=getIcon('mdi6.content-save'), text='Save session')
        btn.clicked.connect(self._onSaveSessionBtnClicked)
        container.layout().addWidget(btn)
        self._saveBtn = btn

        self._saveShortcut = QtWidgets.QShortcut(QtGui.QKeySequence.Save, self._navigatorGUI._win, None, None, QtCore.Qt.ApplicationShortcut)
        self._saveShortcut.activated.connect(self._onSaveSessionShortcutActivated)

        btn = QtWidgets.QPushButton(icon=getIcon('mdi6.content-save-edit-outline'), text='Save session to dir...')
        btn.clicked.connect(lambda checked: self._saveSessionToDir())
        container.layout().addWidget(btn)
        self._saveToDirBtn = btn

        btn = QtWidgets.QPushButton(icon=getIcon('mdi6.content-save-edit'), text='Save session to file...')
        btn.clicked.connect(lambda checked: self._saveSessionToFile())
        container.layout().addWidget(btn)
        self._saveToFileBtn = btn

        container.layout().addSpacing(10)

        btn = QtWidgets.QPushButton(icon=getIcon('mdi6.file-remove'), text='Close session')
        btn.clicked.connect(lambda checked: self._tryVerifyThenCloseSession())
        container.layout().addWidget(btn)
        self._closeBtn = btn

        container.layout().addSpacing(10)

        btn = QtWidgets.QPushButton(icon=getIcon('mdi6.plus'), text='Enable addon')
        btn.clicked.connect(lambda checked: self._addAddon())
        container.layout().addWidget(btn)
        self._addAddonBtn = btn

        container.layout().addStretch()

        title = 'Settings'
        dock = Dock(
            name=self._key + title,
            closable=False,
            title=title,
            affinities=[self._key])
        self._wdgt.addDock(dock, position='right')
        self._infoDW = dock
        container = QtWidgets.QWidget()
        self._settingsContainer = container
        dock.addWidget(container)
        container.setLayout(QtWidgets.QVBoxLayout())

        container = QtWidgets.QGroupBox('Session info')
        self._settingsContainer.layout().addWidget(container)
        container.setLayout(QtWidgets.QFormLayout())
        self._infoContainer = container

        wdgt = QtWidgets.QLineEdit()
        wdgt.setReadOnly(True)
        self._infoWdgts['filepath'] = wdgt
        container.layout().addRow('Session filepath', wdgt)

        wdgt = QtWidgets.QLineEdit()
        wdgt.editingFinished.connect(lambda key='subjectID': self._onInfoTextEdited(key))
        self._infoWdgts['subjectID'] = wdgt
        container.layout().addRow('Subject ID', wdgt)
        # TODO: continue here

        wdgt = QtWidgets.QLineEdit()
        wdgt.editingFinished.connect(lambda key='sessionID': self._onInfoTextEdited(key))
        self._infoWdgts['sessionID'] = wdgt
        container.layout().addRow('Session ID', wdgt)

        container = QtWidgets.QGroupBox('Appearance')
        self._settingsContainer.layout().addWidget(container)
        container.setLayout(QtWidgets.QFormLayout())
        container.layout().setFieldGrowthPolicy(QtWidgets.QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        self._appearanceContainer = container

        wdgt = QtWidgets.QComboBox()
        wdgt.addItems(['Auto', 'Light', 'Dark'])
        wdgt.setCurrentIndex(1)
        wdgt.setMinimumWidth(80)
        wdgt.currentIndexChanged.connect(lambda index: self._onThemeDropdownChanged())
        container.layout().addRow('Theme', wdgt)
        self._themeDropdown = wdgt

        wdgt = QtWidgets.QSpinBox()
        wdgt.setRange(4, 64)
        wdgt.setSingleStep(1)
        wdgt.setMinimumWidth(80)
        wdgt.valueChanged.connect(self._onFontSizeFieldChanged)
        wdgt.clear()
        wdgt.lineEdit().setPlaceholderText('Default')
        container.layout().addRow('Font size', wdgt)
        self._fontSizeField = wdgt

        container = QtWidgets.QGroupBox('About')
        self._settingsContainer.layout().addWidget(container)
        container.setLayout(QtWidgets.QFormLayout())
        container.layout().setFieldGrowthPolicy(QtWidgets.QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)

        wdgt = QtWidgets.QLabel(__version__)
        self._infoWdgts['version'] = wdgt
        container.layout().addRow('NaviNIBS version', wdgt)

        wdgt = QtWidgets.QPushButton('Open NaviNIBS documentation')
        wdgt.clicked.connect(lambda *args: QtGui.QDesktopServices.openUrl('https://precisionneurolab.github.io/navinibs-docs/'))
        container.layout().addRow('Help', wdgt)

        self._settingsContainer.layout().addStretch()

        self._updateEnabledWdgts()

        self._autosaveTask = asyncio.create_task(asyncTryAndLogExceptionOnError(self._autosaveOccasionally))

    def addSaveButtonToDockTabStrip(self, dockArea: DockArea):
        # hack in a save button next to root dock area tab strip

        tCnt = dockArea.topContainer
        if not isinstance(tCnt, TContainer):
            # find shallowest TContainer child

            def findTContainer(child) -> tuple[TContainer | None, int | None]:
                if isinstance(child, TContainer):
                    return child, 0

                if not isinstance(child, pgdc.SplitContainer):
                    return None, None

                tCnt = None
                depth = None
                for i in range(child.count()):
                    tCnt_i, depth_i = findTContainer(child.widget(i))
                    if tCnt_i is not None:
                        if depth is None or depth_i < depth:
                            tCnt = tCnt_i
                            depth = depth_i + 1
                return tCnt, depth

            tCnt, _ = findTContainer(tCnt)

        assert isinstance(tCnt, TContainer)

        if self._rootDockArea is not None:
            assert self._rootDockArea is dockArea
            assert self._tabSaveBtn is not None
            #tCnt.layout.takeAt(tCnt.layout.indexOf(self._tabSaveBtn))
            self._tabSaveBtn = None
        else:
            assert self._tabSaveBtn is None, 'Save button already added to dock tab strip'
            self._rootDockArea = dockArea

        btn = QtWidgets.QPushButton(icon=getIcon('mdi6.content-save'), text='Save')
        btn.clicked.connect(self._onSaveSessionBtnClicked)
        btn.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self._tabSaveBtn = btn

        tCnt.layout.addWidget(btn, 0, 0)  # other widgets in TContainer already start at column index 1

        # update tCnt.stack to span two columns in grid layout
        tCnt.layout.takeAt(tCnt.layout.indexOf(tCnt.stack))
        tCnt.layout.addWidget(tCnt.stack, 1, 0, 1, 2)

        self._updateEnabledWdgts()
        self._updateSaveBtnStyle()

    def _onSessionSet(self):
        self._updateEnabledWdgts()
        if self.session is not None:
            self.session.sigInfoChanged.connect(self._onSessionInfoChanged)
            self.session.sigDirtyKeysChanged.connect(self._updateSaveBtnStyle)
            self.session.miscSettings.sigAttribsChanged.connect(self._onSessionMiscSettingsChanged)
        self._onSessionInfoChanged()
        self._onSessionMiscSettingsChanged()

    def _onThemeDropdownChanged(self):
        theme = self._themeDropdown.currentText()
        self.session.miscSettings.theme = theme.lower()

    def _onFontSizeFieldChanged(self, *args):
        fontSize = self._fontSizeField.value()
        self.session.miscSettings.mainFontSize = fontSize

    def _getNewInProgressSessionDir(self, suffix: str = '') -> str:
        return os.path.join(self._inProgressBaseDir, 'NaviNIBSSession_' + suffix + datetime.today().strftime('%y%m%d%H%M%S'))

    def _updateEnabledWdgts(self):
        wdgts = [self._saveBtn, self._saveToFileBtn, self._saveToDirBtn,
                 self._importBtn,
                 self._closeBtn, self._addAddonBtn,
                 self._infoContainer,
                 self._appearanceContainer]
        if self._tabSaveBtn is not None:
            wdgts.append(self._tabSaveBtn)
        for wdgt in wdgts:
            try:
                wdgt.setEnabled(self.session is not None)
            except Exception as e:
                pass  # widget may have been deleted

    def _updateSaveBtnStyle(self):
        btn = self._tabSaveBtn
        if btn is None:
            return
        isDirty = self.session is not None and len(self.session.dirtyKeys) > 0
        if self._saveBtnShowsDirty is not None and isDirty == self._saveBtnShowsDirty:
            return
        self._saveBtnShowsDirty = isDirty
        palette = self.wdgt.palette()
        if isDirty:
            fg = palette.color(QtGui.QPalette.Active, QtGui.QPalette.Text).name()
        else:
            fg = palette.color(QtGui.QPalette.Disabled, QtGui.QPalette.Text).name()
        try:
            btn.setStyleSheet(f"""
                    QPushButton {{
                        border: none;
                        color: {fg};
                    }}
                    QPushButton:hover {{background-color: rgba(0, 0, 0, 0.1);}}
                    QPushButton:pressed {{background-color: rgba(0, 0, 0, 0.2);}}
                """)
        except Exception as e:
            # button may have been deleted
            pass

    def _updateLayoutsBeforeSave(self):
        """
        We don't update serialized layouts on autosaves or other frequent GUI updates, but do want to update
        right before a manual session save.
        """
        self._navigatorGUI.saveLayout()

    def _onSaveSessionBtnClicked(self, checked: bool):
        # if 'alt' modifier is pressed during click, then force save all (ignoring dirty flags)
        self._updateLayoutsBeforeSave()
        if QtWidgets.QApplication.keyboardModifiers() == QtCore.Qt.AltModifier:
            self.session.saveToFile(updateDirtyOnly=False)
        else:
            self.session.saveToFile()

    def _onSaveSessionShortcutActivated(self):
        self._updateLayoutsBeforeSave()
        if QtWidgets.QApplication.keyboardModifiers() == QtCore.Qt.AltModifier:
            self.session.saveToFile(updateDirtyOnly=False)
        else:
            self.session.saveToFile()

    def _saveSessionToFile(self, sesFilepath: tp.Optional[str] = None):
        if sesFilepath is None:
            prevFilepath = self.session.filepath
            sesFilepath, _ = QtWidgets.QFileDialog.getSaveFileName(self._wdgt,
                                                                   'Save session to file',
                                                                   prevFilepath,
                                                                   'Session file (*.navinibs)')
            if len(sesFilepath) == 0:
                logger.info('Browse save session file cancelled')
                return
        logger.info('New session filepath: {}'.format(sesFilepath))
        self.session.filepath = sesFilepath
        self._updateLayoutsBeforeSave()
        self.session.saveToFile()

    def _saveSessionToDir(self, sesFilepath: tp.Optional[str] = None):
        if sesFilepath is None:
            prevFilepath = self.session.filepath
            sesFilepath = QtWidgets.QFileDialog.getExistingDirectory(self._wdgt,
                                                                   'Save session to dir',
                                                                   prevFilepath)
            if len(sesFilepath) == 0:
                logger.info('Browse save session dir cancelled')
                return
        logger.info('New session filepath: {}'.format(sesFilepath))
        self.session.filepath = sesFilepath
        self._updateLayoutsBeforeSave()
        self.session.unpackedSessionDir = sesFilepath  # this will trigger copy to new destination
        self.session.saveToFile()

    def _closeSession(self):
        closedSession = self.session
        self.session = None
        self.sigClosedSession.emit(closedSession)

    def _tryVerifyThenCloseSession(self):
        if self.session is not None:
            if self.session.compressedFileIsDirty:
                raise NotImplementedError()  # TODO: prompt user to confirm whether they want to save or discard changes to previous session
                # TODO: close session to trigger clearing of various GUI components
            self._closeSession()

    def _addAddon(self, addonFilepath: str | None = None):
        if addonFilepath is None:
            addonFilepath, _ = QtWidgets.QFileDialog.getOpenFileName(
                self._wdgt,
                'Select addon configuration file',
                addonBaseInstallPath,
                'Addon configuration file (*.json)')
            if len(addonFilepath) == 0:
                logger.info('Browse addon file cancelled')
                return

        logger.info('Selected addon file: {}'.format(addonFilepath))

        addonDict = dict()
        addonDirPath = os.path.dirname(addonFilepath)
        addonDict['addonInstallPath'] = os.path.relpath(addonDirPath, addonBaseInstallPath)

        addon = Addon.fromDict(addonDict)

        self.session.addons.addItem(addon)

        logger.info(f'Added addon {addon.key}')

    def _createNewSession(self, sesFilepath: tp.Optional[str] = None):
        self._tryVerifyThenCloseSession()

        if sesFilepath is None:
            if False:
                raise NotImplementedError()  # TODO: set to location of recent dir if available
                dir = 'todo'
            else:
                dir = str(pathlib.Path.home())


            useFilters = False

            if useFilters:
                sesFilepath, selectedFilter = QtWidgets.QFileDialog.getSaveFileName(self._wdgt,
                                                                    'Create new session file',
                                                                    dir,
                                                                    'Session file (*.navinibs);;Session folder (*)',
                                                                    'Session folder (*)')
            else:
                sesFilepath, selectedFilter = QtWidgets.QFileDialog.getSaveFileName(self._wdgt,
                                                                                    'Create new session file',
                                                                                    dir)

            if len(sesFilepath) == 0:
                logger.info('Browse new session cancelled')
                return

            if useFilters:
                # due to some issue with QFileDialog, if ext filters are specified, then an extension is always
                # appended, even if none was specified and this empty ext matches an existing filter.
                # So strip extension manually if the empty filter was selected.
                # This has the side effect that the .navinibs filter MUST be selected to specify a .navinibs path.
                if selectedFilter == 'Session folder (*)':
                    sesFilepath = os.path.splitext(sesFilepath)[0]

        logger.info('New session filepath: {}'.format(sesFilepath))

        if sesFilepath.endswith('.navinibs'):
            unpackedSessionDir = self._getNewInProgressSessionDir()
        else:
            unpackedSessionDir = sesFilepath
        session = Session.createNew(filepath=sesFilepath, unpackedSessionDir=unpackedSessionDir)
        self.sigAboutToFinishLoadingSession.emit(session)
        self.session = session
        self.sigLoadedSession.emit(self.session)

    def _loadSession(self, sesFilepath: str, unpackedSessionSuffix: str = '') -> Session | None:
        if sesFilepath is None:
            if self.session is not None:
                dir = os.path.dirname(self.session.filepath)
            else:
                if False:
                    raise NotImplementedError()  # TODO: set to location of recent dir if available
                    dir = 'todo'
                else:
                    dir = str(pathlib.Path.home())
            sesFilepath, _ = QtWidgets.QFileDialog.getOpenFileName(self._wdgt, 'Choose session to load', dir, 'Session file (*.navinibs); Config file (*.json)')
            if len(sesFilepath) == 0:
                logger.info('Browse existing session cancelled')
                return None

        sesFilepath = os.path.normpath(sesFilepath)

        logger.info('Load session filepath: {}'.format(sesFilepath))

        if sesFilepath.endswith('.json'):
            # assume this is a config file inside a larger session dir
            sesFilepath = os.path.dirname(sesFilepath)

        try:
            if os.path.isdir(sesFilepath):
                # treat as already-unpacked session dir
                session = Session.loadFromFolder(folderpath=sesFilepath)
            else:
                # treat as compressed file
                assert os.path.isfile(sesFilepath)
                session = Session.loadFromFile(filepath=sesFilepath, unpackedSessionDir=self._getNewInProgressSessionDir(suffix=unpackedSessionSuffix))
        except Exception as e:
            raiseErrorDialog(f'Problem loading session from {sesFilepath}', exception=e)
            return None

        return session

    def loadSession(self, sesFilepath: tp.Optional[str] = None):
        self._tryVerifyThenCloseSession()

        session = self._loadSession(sesFilepath=sesFilepath)

        if session is None:
            return

        self.sigAboutToFinishLoadingSession.emit(session)
        self.session = session
        try:
            self.sigLoadedSession.emit(self.session)
        except Exception as e:
            logger.error('Problem handling loaded session:\n{}'.format(exceptionToStr(e)))
            raise e

    def importSession(self, sesFilepath: tp.Optional[str] = None):
        logger.info(f'Loading other session for import from {sesFilepath}')
        otherSession = self._loadSession(sesFilepath=sesFilepath, unpackedSessionSuffix='TmpImport')
        assert self._importSessionWindow is None
        self._importSessionWindow = ImportSessionWindow(
            parent=self._navigatorGUI._win,
            session=self.session,
            otherSession=otherSession
        )
        self._importSessionWindow.sigFinished.connect(self._onImportWindowFinished)
        self._importSessionWindow.show()

    def _onImportWindowFinished(self, didAccept: bool):
        assert self._importSessionWindow is not None
        self._importSessionWindow.sigFinished.disconnect(self._onImportWindowFinished)
        self._importSessionWindow = None
        logger.info('Import session window closed')

    def _recoverSession(self, sesDataDir: tp.Optional[str] = None):
        self._tryVerifyThenCloseSession()

        if sesDataDir is None:
            dir = self._inProgressBaseDir
            sesDataDir = QtWidgets.QFileDialog.getExistingDirectory(self._wdgt, 'Choose unpacked session to load', dir)
            if len(sesDataDir) == 0:
                logger.info('Browse recover session cancelled')
                return
        logger.info('Recover session data dir: {}'.format(sesDataDir))

        session = Session.loadFromUnpackedDir(unpackedSessionDir=sesDataDir)
        self.sigAboutToFinishLoadingSession.emit(session)
        self.session = session
        self.sigLoadedSession.emit(self.session)

    def _cloneSession(self, fromSesFilepath: tp.Optional[str] = None, toSesFilepath: tp.Optional[str] = None):
        if fromSesFilepath is None:
            if False:
                raise NotImplementedError()  # TODO: set to location of recent dir if available
                dir = 'todo'
            else:
                dir = str(pathlib.Path.home())
            fromSesFilepath, _ = QtWidgets.QFileDialog.getOpenFileName(self._wdgt, 'Choose session to clone', dir,
                                                                "Session file (*.navinibs)")
            if len(sesFilepath) == 0:
                logger.info('Browse existing session cancelled')
                return

        fromSesFilepath = os.path.normpath(fromSesFilepath)

        logger.info('Load session filepath: {}'.format(fromSesFilepath))

        if toSesFilepath is None:
            dir, _ = os.path.split(fromSesFilepath)
            toSesFilepath, _ = QtWidgets.QFileDialog.getSaveFileName(self._wdgt, 'Create save cloned session file', dir, "Session file (*.navinibs)")
            if len(sesFilepath) == 0:
                logger.info('Browse clone session cancelled')
                return

        toSesFilepath = os.path.normpath(toSesFilepath)

        logger.info('Cloned session filepath: {}'.format(toSesFilepath))

        logger.debug('Copying session from {} to {}'.format(fromSesFilepath, toSesFilepath))
        shutil.copyfile(fromSesFilepath, toSesFilepath)
        logger.debug('Done copying')

        self.loadSession(sesFilepath=toSesFilepath)

    def _onSessionInfoChanged(self, whatChanged: tp.Optional[list[str]] = None):
        allRelevantKeys = ('filepath', 'subjectID', 'sessionID')
        if whatChanged is None:
            whatChanged = allRelevantKeys
        else:
            whatChanged = tuple(key for key in whatChanged if key in allRelevantKeys)
        if self.session is None:
            for key in whatChanged:
                self._infoWdgts[key].setText('')
        else:
            for key in whatChanged:
                val = getattr(self.session, key)
                self._infoWdgts[key].setText('' if val is None else val)

    def _onSessionMiscSettingsChanged(self, whatChanged: tp.Optional[list[str]] = None):
        if whatChanged is None or 'theme' in whatChanged:
            theme = self.session.miscSettings.theme.capitalize()
            self._themeDropdown.setCurrentText(theme)

        if whatChanged is None or 'mainFontSize' in whatChanged:
            fontSize = self.session.miscSettings.mainFontSize
            if fontSize is None:
                self._fontSizeField.clear()
            else:
                self._fontSizeField.setValue(fontSize)

    def _onInfoTextEdited(self, key: str):
        text = self._infoWdgts[key].text()
        if len(text) == 0:
            text = None
        if self.session is not None:
            logger.info('Applying edited value of {} to session: {}'.format(key, text))
            setattr(self.session, key, text)
        else:
            logger.warning('Ignoring edited value of {} since session is closed.'.format(key))

    async def _autosave(self):
        logger.debug('Trying to autosave')
        if self.session is None:
            logger.debug('Session is None, nothing to autosave.')
        else:
            self.session.saveToUnpackedDir(asAutosave=True)
        logger.debug('Done trying to autosave')

    async def _autosaveOccasionally(self):
        while True:
            await asyncio.sleep(self._autosavePeriod)
            await self._autosave()

    def restoreLayoutIfAvailable(self) -> bool:
        if not super().restoreLayoutIfAvailable():
            return False

        if self._rootDockArea is not None:
            self.addSaveButtonToDockTabStrip(self._rootDockArea)

        return True


from __future__ import annotations

import keyboard
import typing as tp

import attrs
import logging
from qtpy import QtWidgets

from .TriggerSourceSettingsWidget import TriggerSourceSettingsWidget
from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.Navigator.Model.Triggering import HotkeyTriggerSource, Hotkey, Hotkeys, TriggerEvent
from NaviNIBS.Navigator.GUI.CollectionModels.HotkeysTableModel import HotkeysTableModel
from NaviNIBS.Navigator.GUI.Widgets.CollectionTableWidget import CollectionTableWidget


logger = logging.getLogger(__name__)


@attrs.define(kw_only=True)
class HotkeysTableWidget(CollectionTableWidget[str, Hotkey, Hotkeys, HotkeysTableModel]):
    _triggerSourceKey: str
    _Model: tp.Callable[[Session], HotkeysTableModel] = None

    def __attrs_post_init__(self):
        self._Model = lambda session: HotkeysTableModel(session=session, triggerSourceKey=self._triggerSourceKey)

        super().__attrs_post_init__()


@attrs.define
class HotkeyTriggerSourceSettingsWidget(TriggerSourceSettingsWidget[HotkeyTriggerSource]):
    _triggerSourceKey: str
    _title: str = 'Hotkey trigger settings'

    _hotkeyCallbacks: dict[str, tp.Callable] = attrs.field(init=False, factory=dict)
    _hotkeysTable: HotkeysTableWidget = attrs.field(init=False)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._hotkeysTable = HotkeysTableWidget(triggerSourceKey=self._triggerSourceKey, session=self.session)
        self._wdgt.setLayout(QtWidgets.QHBoxLayout())
        self._wdgt.layout().addWidget(self._hotkeysTable.wdgt)

        self.triggerSource.hotkeys.sigItemsAboutToChange.connect(self._onHotkeysAboutToChange)
        self.triggerSource.hotkeys.sigItemsChanged.connect(self._onHotkeysChanged)
        self.triggerSource.hotkeys.sigItemKeyAboutToChange.connect(self._onHotkeyKeyAboutToChange)
        self.triggerSource.hotkeys.sigItemKeyChanged.connect(self._onHotkeyKeyChanged)

        for key in self.triggerSource.hotkeys.keys():
            self._registerHotkey(key)

    def _onHotkeysAboutToChange(self, keys: list[str], attribs: tp.Optional[list[str]]):
        if attribs is None:
            # key may be being removed
            for key in keys:
                if key in self._hotkeyCallbacks:
                    self._unregisterHotkey(key)

    def _onHotkeysChanged(self, keys: list[str], attribs: tp.Optional[list[str]]):
        if attribs is None:
            # key may have just been removed or added
            for key in keys:
                if key in self.triggerSource.hotkeys:
                    # key may have just been added
                    self._registerHotkey(key)
                else:
                    pass # key was just removed, already handled by aboutToChange above

    def _onHotkeyKeyAboutToChange(self, oldKey: str, newKey: str):
        if oldKey in self._hotkeyCallbacks:
            self._unregisterHotkey(oldKey)

    def _onHotkeyKeyChanged(self, oldKey: str, newKey: str):
        pass  # wil be registered by onHotkeysChanged

    def _onKeyPressedStage1(self, whichKey, evt: keyboard.KeyboardEvent) -> bool:
        logger.debug(f'Key pressed stage 1: {whichKey} {evt}')
        return True  # don't block

    def _onKeyPressed(self, whichKey: str, evt: keyboard.KeyboardEvent) -> bool:
        hotkey = self.triggerSource.hotkeys[whichKey]
        if hotkey.keyboardDeviceID is None or str(evt.device) == str(hotkey.keyboardDeviceID):
            logger.info(f'Hotkey pressed: {whichKey} ({evt})')
            triggerEvt = TriggerEvent(type=hotkey.action)
            self.triggerSource.trigger(triggerEvt)
        else:
            logger.debug(f'Hotkey pressed but did not match keyboardDeviceID: {whichKey} {evt}')
        return True  # don't block

    def _registerHotkey(self, key: str):
        logger.info(f'Registering hotkey {key}')
        assert key not in self._hotkeyCallbacks
        hotkey = self.triggerSource.hotkeys[key]
        if True:
            keyToHook = key
        else:
            keyToHook = keyboard.key_to_scan_codes(key)[-1]

        if True:
            callback1 = lambda evt, key=key: self._onKeyPressedStage1(whichKey=key, evt=evt)
            callback2 = lambda evt, key=key: self._onKeyPressed(whichKey=key, evt=evt)
            callback = (callback1, callback2)
            kwargs = dict(suppress=True)
        else:
            callback = lambda evt, key=key: self._onKeyPressed(whichKey=key, evt=evt)
            kwargs = dict()

        self._hotkeyCallbacks[key] = callback

        keyboard.hook_key(keyToHook, callback=callback, **kwargs)

    def _unregisterHotkey(self, key: str):
        logger.info(f'Unregistering hotkey {key}')
        assert key in self._hotkeyCallbacks
        keyboard.unhook(self._hotkeyCallbacks.pop(key))



import attrs
import typing as tp


def attrsAsDict(obj,
                eqs: tp.Optional[tp.Dict[str, tp.Callable[[tp.Any, tp.Any], bool]]] = None,
                exclude: tp.Optional[tp.Iterable[str]] = None) -> tp.Dict[str, tp.Any]:
    """
    Similar to attrs.asdict(), but strips underscore from private attributes, doesn't include items with unchanged defaults, etc.

    Specify eqs to provide custom equality test callables, e.g.
     dict(field1=array_equalish, field2=nested_array_equalish) for handling np.ndarrays and dicts with ndarray values, respectively
    """

    def filterAttrs(attrib: attrs.Attribute, val: tp.Any,
                    eqs: tp.Optional[tp.Dict[str, tp.Callable[[tp.Any, tp.Any], bool]]] = None,
                    exclude: tp.Optional[tp.Iterable[str]] = None):
        if not attrib.init:
            # don't include non-init attribs
            return False

        if exclude is not None and (attrib.name in exclude or ('_' + attrib.name) in exclude or
                (attrib.name.startswith('_') and attrib.name[1:] in exclude)):
            # don't include excluded attribs
            return False

        if True:
            # don't include values that still equal default value

            # noinspection PyTypeChecker
            if isinstance(attrib.default, attrs.Factory):
                if attrib.default.takes_self:
                    raise NotImplementedError()  # TODO:
                else:
                    default = attrib.default.factory()
            else:
                default = attrib.default

            if attrib.default is attrs.NOTHING:
                return True  # no default set, so include value without checking equality

            if attrib.eq:
                if eqs is not None:
                    if attrib.name not in eqs and attrib.name.startswith('_'):
                        attrib_name_eqs = attrib.name[1:]
                    else:
                        attrib_name_eqs = attrib.name
                else:
                    attrib_name_eqs = None
                if attrib_name_eqs is not None and attrib_name_eqs in eqs:
                    eq = eqs[attrib_name_eqs](val, default)
                elif attrib.eq:
                    eq = default == val
                else:
                    eq = default is val
                if eq:
                    return False

        return True

    d = attrs.asdict(obj, filter=lambda attrib, val, eqs=eqs, exclude=exclude: filterAttrs(
        attrib, val, eqs=eqs, exclude=exclude))

    # remove underscores before private attributes
    d = {(key[1:] if key.startswith('_') else key): val for key, val in d.items()}

    return d








import asyncio
import logging
from functools import wraps
import typing as tp

from . import exceptionToStr
from .AsyncRunner import asyncioRunAndHandleExceptions

logger = logging.getLogger(__name__)

_T = tp.TypeVar('_T')


async def _wrap_awaitable(aw: tp.Awaitable[_T]) -> _T:
    return await aw


async def asyncWait(
        aws: tp.Iterable[tp.Awaitable],
        return_when=asyncio.ALL_COMPLETED) -> tuple[set[asyncio.Task], set[asyncio.Task]]:
    """
    Similar to asyncio.wait, but allow waiting for coroutines, which was removed from asyncio.wait in 3.11
    """
    done, pending = await asyncio.wait([asyncio.create_task(_wrap_awaitable(aw)) for aw in aws], return_when=return_when)
    return done, pending


async def asyncWaitWithCancel(
        aws: tp.Iterable[tp.Awaitable],
        timeout: float | None = None,
        return_when=asyncio.ALL_COMPLETED) -> tuple[set[asyncio.Task], set[asyncio.Task]]:

    done, pending = await asyncio.wait([asyncio.create_task(_wrap_awaitable(aw)) for aw in aws], timeout=timeout, return_when=return_when)
    for task in pending:
        task.cancel()
    cancelled = pending
    return done, cancelled


async def asyncTryAndLogExceptionOnError(fn: tp.Callable[..., tp.Awaitable], *args, **kwargs):
    try:
        return await fn(*args, **kwargs)
    except Exception as e:
        logger.error('Exception: %s' % exceptionToStr(e))
        raise e


def asyncAtomicCancellable(fn: tp.Callable[..., tp.Awaitable], *args, **kwargs):
    @wraps(fn)
    async def wrapper(*args, **kwargs):
        task = asyncio.create_task(fn(*args, **kwargs))
        shieldedTask = asyncio.shield(task)
        try:
            return await shieldedTask
        except asyncio.CancelledError:
            await task
            raise
    return wrapper


def asyncNonCancellable(fn: tp.Callable[..., tp.Awaitable], *args, **kwargs):
    @wraps(fn)
    async def wrapper(*args, **kwargs):
        task = asyncio.create_task(fn(*args, **kwargs))
        shieldedTask = asyncio.shield(task)
        try:
            return await shieldedTask
        except asyncio.CancelledError:
            return await task
    return wrapper




from qtpy import QtWidgets


def setStyleSheetForInstanceOnly(instance: QtWidgets.QWidget, styleSheet: str,
                                 selectorPrefix: str = '',
                                 selectorSuffix: str = ''):
    """
    Set style sheet for instance only, not for its children.

    Adapted from https://stackoverflow.com/a/70705685/2388228
    """
    objectName = instance.objectName()
    if len(objectName) == 0:
        objectName = str(id(instance))
        instance.setObjectName(objectName)
    instance.setStyleSheet(f'{selectorPrefix}#{objectName}{selectorSuffix} {{{styleSheet}}}')



import attrs
from qtpy import QtWidgets, QtCore, QtGui


@attrs.define
class QCollapsibleSection:
    _title: str
    _innerWdgt: QtWidgets.QWidget = attrs.field(factory=QtWidgets.QWidget)
    _animationDuration: int = 300  # in ms
    _doStartCollapsed: bool = False

    _button: QtWidgets.QToolButton = attrs.field(init=False, factory=QtWidgets.QToolButton)
    _outerWdgt: QtWidgets.QWidget = attrs.field(init=False, factory=QtWidgets.QWidget)
    _animation: QtCore.QPropertyAnimation = attrs.field(init=False)

    def __attrs_post_init__(self):
        layout = QtWidgets.QVBoxLayout()
        self._outerWdgt.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)
        self._button.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        self._button.setIconSize(QtCore.QSize(8, 8))
        self._button.setText(" " + self._title)
        self._button.setCheckable(True)
        self._button.setStyleSheet("""
        QToolButton {
            background: none;
            border: none
            
        }
        QToolButton:pressed {
            background: none;
            border: none
        }   
        """)
        self._button.setChecked(not self._doStartCollapsed)
        if self._doStartCollapsed:
            self._innerWdgt.setMaximumHeight(0)

        self._button.setArrowType(QtCore.Qt.DownArrow if self._button.isChecked() else QtCore.Qt.RightArrow)

        self._button.toggled.connect(self._onButtonToggled)

        layout.addWidget(self._button)
        layout.addWidget(self._innerWdgt)

        animation = QtCore.QPropertyAnimation(self._innerWdgt, b"maximumHeight")
        animation.setStartValue(0)
        animation.setEasingCurve(QtCore.QEasingCurve.InOutCubic)
        animation.setDuration(self._animationDuration)  # TODO: change to shorter duration by default
        animation.setEndValue(self._innerWdgt.geometry().height() + 10)
        self._animation = animation

    @property
    def outerWdgt(self):
        return self._outerWdgt

    @property
    def innerWdgt(self):
        return self._innerWdgt

    def setLayout(self, layout: QtWidgets.QLayout):
        """
        Set layout of innerWdgt
        """
        self._innerWdgt.setLayout(layout)

    def _onButtonToggled(self, checked: bool):
        self._button.setArrowType(QtCore.Qt.DownArrow if checked else QtCore.Qt.RightArrow)
        if checked:
            self.showContent()
        else:
            self.hideContent()

    def showContent(self):
        self._animation.setDirection(QtCore.QAbstractAnimation.Forward)
        self._animation.setEndValue(self._innerWdgt.sizeHint().height() + 10)
        self._animation.start()

    def hideContent(self):
        self._animation.setDirection(QtCore.QAbstractAnimation.Backward)
        self._animation.setEndValue(self._innerWdgt.sizeHint().height() + 10)
        self._animation.start()

